
import xml.etree.ElementTree as ET
import glob, os, sys, shutil
from datetime import datetime, timedelta
try:
    from urllib2 import urlopen, URLError
    from urllib import quote
    from urlparse import urlparse, urlunsplit
except ImportError:
    from urllib.request import urlopen
    from urllib.error import URLError
    from urllib.parse import quote, urlparse, urlunsplit

import certifi
import platform, subprocess, re
import imp
import socket
import logging
import traceback

import sharppy.io.decoder as decoder
from sharppy.io.csv import loadCSV
from sharppy.io.csv import loadNUCAPS_CSV

import sutils.frozenutils as frozenutils

HOME_DIR = os.path.join(os.path.expanduser("~"), ".sharppy", "datasources")
NUCAPS_times_file = os.path.join(HOME_DIR, "nucapsTimes.txt") # JTS

# Cleanup; Remove nucapsTimes.txt and cloudTopValues.txt when main GUI opens.
if os.path.isfile(NUCAPS_times_file):
    os.remove(NUCAPS_times_file)

if not os.path.exists(HOME_DIR):
    os.makedirs(HOME_DIR)

if frozenutils.isFrozen():
    from . import available
else:
    avail_loc = os.path.join(HOME_DIR, 'available.py')
    if not os.path.exists(avail_loc):
        pkg_avail_loc = os.path.join(os.path.dirname(__file__), 'available.py')
        shutil.copy(pkg_avail_loc, avail_loc)

    available = imp.load_source('available', avail_loc)

# TAS: Comment this file and available.py

class DataSourceError(Exception):
    pass

def loadDataSources(ds_dir=HOME_DIR):
    """
    Load the data source information from the XML files.
    Returns a dictionary associating data source names to DataSource objects.
    """

    files = glob.glob(os.path.join(ds_dir, '*.xml'))
    if len(files) == 0:
        if frozenutils.isFrozen():
            frozen_path = frozenutils.frozenPath()
            files = glob.glob(os.path.join(frozen_path, 'sharppy', 'datasources', '*.xml')) +  \
                    glob.glob(os.path.join(frozen_path, 'sharppy', 'datasources', '*.csv'))
        else:
            pkg_path = os.path.dirname(__file__)
            files = glob.glob(os.path.join(pkg_path, '*.xml')) + glob.glob(os.path.join(pkg_path, '*.csv'))

        for file_name in files:
            shutil.copy(file_name, ds_dir)

    files = glob.glob(os.path.join(ds_dir, '*.xml'))
    ds = {}
    for ds_file in files:
        root = ET.parse(ds_file).getroot()
        for src in root:
            name = src.get('name')
            try:
                ds[name] = DataSource(src)
            except Exception as e:
                traceback.print_exc()
                print("Exception: ", e)
                print('Unable to process %s file'%os.path.basename(ds_file))
                print("This data source may not be loaded then.")
    return ds

def _pingURL(hostname, timeout=1):
    try:
        urlopen(hostname, timeout=timeout, cafile=certifi.where())
    except URLError:
        return False
    except socket.timeout as e:
        return False

    return True

def pingURLs(ds_dict):
    """
    Try to ping all the URLs in any XML file.
    Takes a dictionary associating data source names to DataSource objects.
    Returns a dictionary associating URLs with a boolean specifying whether or not they were reachable.
    """
    urls = {}

    for ds in list(ds_dict.values()):
        ds_urls = ds.getURLList()
        for url in ds_urls:
            urlp = urlparse(url)
            base_url = urlunsplit((urlp.scheme, urlp.netloc, '', '', ''))
            urls[base_url] = None

    for url in urls.keys():
        urls[url] = _pingURL(url)

        if urls[url]:
            # A bit of a hack. Since we're only using this to check for an Internet connection, we don't need to check
            # any more if we find one.
            break
    return urls

class Outlet(object):
    """
    An Outlet object contains all the information for a data outlet (for example, the observed
    sounding feed from SPC, which is distinct from the observed sounding feed from, say, Europe).
    """
    def __init__(self, ds_name, config):
        self._ds_name = ds_name
        self._name = config.get('name')
        self._url = config.get('url')
        self._format = config.get('format')
        self._time = config.find('time')
        point_csv = config.find('points')
        #self.start, self.end = self.getTimeSpan()

        # JTS - Parse remote CSVs from FTP site for NUCAPS data sources.
        # Read the NUCAPS data source, region, satellite ID, year, month, day and time info
        # from the temporary file and assign to local variables.
        if os.path.isfile(NUCAPS_times_file):
            file = open(NUCAPS_times_file)
            line = file.readlines()

            # Remove the list surrounding the values.
            line = line[0]
            selected_ds = str(line.split(',')[0])

            # If data source in xml matches selected data source, then grab the associated remote csv file.
            if self._ds_name == selected_ds:
                region = str(line.split(',')[1])
                satID = str(line.split(',')[2])
                YEAR = str(line.split(',')[3])
                MONTH = str(line.split(',')[4])
                DAY = str(line.split(',')[5])
                TIME = str(line.split(',')[6])

                remote_csv = f'https://geo.nsstc.nasa.gov/SPoRT/jpss-pg/nucaps/gridded/{region}/sharppy/{satID}/csv/{YEAR}{MONTH}{DAY}{TIME}/{satID}_{region}.csv'
                self._csv_fields, self._points = loadNUCAPS_CSV(remote_csv)
            else:
                # Do nothing if data source name in xml does not match the selected data source.
                self._csv_fields, self._points = loadCSV(os.path.join(HOME_DIR, point_csv.get("csv")))
        else:
            # Parse local CSVs stored in ~/.sharppy/datasources for non-NUCAPS data sources.
            self._csv_fields, self._points = loadCSV(os.path.join(HOME_DIR, point_csv.get("csv")))

        for idx in range(len(self._points)):
            self._points[idx]['lat'] = float(self._points[idx]['lat'])
            self._points[idx]['lon'] = float(self._points[idx]['lon'])
            self._points[idx]['elev'] = int(self._points[idx]['elev'])

        self._custom_avail = self._name.lower() in available.available and self._ds_name.lower() in available.available[self._name.lower()]
        self._is_available = True

    def getTimeSpan(self):
        """
        Read in the XML data to determine the dates/times this outlet spans across
        """
        try:
            start = self._time.get('start')
        except:
            start = '-'
        try:
            end = self._time.get('end')
        except:
            end = '-'
        if start == "-" or start is None:
            s = None
        elif start == "now":
            s = datetime.utcnow()
        else:
            s = datetime.strptime(start, '%Y/%m/%d')

        if end == "-" or end is None:
            e = None
        elif end == 'now':
            e = datetime.utcnow()
        else:
            e = datetime.strptime(end, '%Y/%m/%d')
        return [s, e]

    def getForecastHours(self):
        """
        Return a list of forecast hours that this outlet serves. If the 'delta' attribute is set to 0,
        then the data source only has the 0-hour.
        """
        times = []
        t = self._time
        f_start = int(t.get('first'))
        f_range = int(t.get('range'))
        f_delta = int(t.get('delta'))
        if f_delta > 0:
            times.extend(list(range(f_start, f_range + f_delta, f_delta)))
        else:
            times.append(0)
        return times

    def getCycles(self):
        """
        Return a list of data cycles that this outlet serves.
        """
        times = []

        t = self._time
        c_length = int(t.get('cycle'))
        c_offset = int(t.get('offset'))
        return [ t + c_offset for t in range(0, 24, c_length) ]

    def getDelay(self):
        return int(self._time.get('delay'))

    def getMostRecentCycle(self, dt=None):
        custom_failed = False

        if self._custom_avail:
            try:
                try:
                    times = available.available[self._name.lower()][self._ds_name.lower()](dt)
                except TypeError:
                    times = available.available[self._name.lower()][self._ds_name.lower()]()
                recent = max(times)
                self._is_available = True
            except URLError:
                custom_failed = True
                self._is_available = False

        if not self._custom_avail or custom_failed:
            now = datetime.utcnow()
            cycles = self.getCycles()
            delay = self.getDelay()
            avail = [ now.replace(hour=hr, minute=0, second=0, microsecond=0) + timedelta(hours=delay) for hr in cycles ]
            avail = [ run - timedelta(days=1) if run > now else run for run in avail ]
            recent = max(avail) - timedelta(hours=delay)
        return recent

    def getArchivedCycles(self, **kwargs):
        max_cycles = kwargs.get('max_cycles', 100)

        start = kwargs.get('start', None)
        if start is None:
            start = self.getMostRecentCycle()

        daily_cycles = self.getCycles()
        time_counter = daily_cycles.index(start.hour)
        archive_len = self.getArchiveLen()
        # TODO: Potentially include the option to specify the beginning date of the archive, if the data source
        # is static?

        cycles = []
        cur_time = start
        while cur_time > start - timedelta(hours=archive_len):
            cycles.append(cur_time)

            if len(cycles) >= max_cycles:
                break

            time_counter = (time_counter - 1) % len(daily_cycles)
            cycle = daily_cycles[time_counter]
            cur_time = cur_time.replace(hour=cycle)
            if cycle == daily_cycles[-1]:
                cur_time -= timedelta(days=1)
        return cycles

    def getAvailableAtTime(self, **kwargs):
        dt = kwargs.get('dt', None)

        logging.debug("Calling getAvailableAtTime()" + str(dt) + " " + self._ds_name.lower() + " " + self._name.lower())
        if dt is None:
            dt = self.getMostRecentCycle(dt)
        elif dt == datetime(1700,1,1,0,0,0):
            return []
        stns_avail = self.getPoints()
        if self._name.lower() in available.availableat and self._ds_name.lower() in available.availableat[self._name.lower()]:
            #avail = available.availableat[self._name.lower()][self._ds_name.lower()](dt)
            try:
                avail = available.availableat[self._name.lower()][self._ds_name.lower()](dt)
                stns_avail = []
                points = self.getPoints()
                srcids = [ p['srcid'] for p in points ]
                for stn in avail:
                    try:
                        idx = srcids.index(stn)
                        stns_avail.append(points[idx])
                    except ValueError:
                        pass

                self._is_available = True

            except URLError as err:
                logging.exception(err)
                stns_avail = []
                self._is_available = False
        logging.debug("_is_available: "+ str(self._is_available))
        logging.debug('Number of stations found:' + str(len(stns_avail)))
        return stns_avail

    def getAvailableTimes(self, max_cycles=100, **kwargs):
        # THIS IS WHERE I COULD POTENTIALLY PASS AN ARGUMENT TO FILTER OUT WHAT RAOBS OR MODELS ARE AVAILABLE
        custom_failed = False
        dt = kwargs.get('dt', None) #Used to help search for times
        logging.debug("Called outlet.getAvailableTimes():" + " dt=" + str(dt) + " self._name=" + str(self._name) + " self._ds_name=" + str(self._ds_name))
        if self._custom_avail:
            try:
                if self._name.lower() == "local":
                    times = available.available[self._name.lower()][self._ds_name.lower()](kwargs.get("filename"))
                else:
                    try:
                        times = available.available[self._name.lower()][self._ds_name.lower()](dt)
                    except TypeError:
                        times = available.available[self._name.lower()][self._ds_name.lower()]()
                if len(times) == 1:
                    times = self.getArchivedCycles(start=times[0], max_cycles=max_cycles)
                self._is_available = True
            except URLError as err:
                logging.exception(err)
                custom_failed = True
                self._is_available = False

        if not self._custom_avail or custom_failed:
            times = self.getArchivedCycles(max_cycles=max_cycles)
        logging.debug("outlet.getAvailableTimes() times Found:" + str(times[~max_cycles:]))
        return times[-max_cycles:]

    def getArchiveLen(self):
        return int(self._time.get('archive'))

    def getURL(self):
        return self._url

    def getDecoder(self):
        return decoder.getDecoder(self._format)

    def hasProfile(self, point, cycle):
        logging.debug("Calling outlet.hasProfile() ")
        times = self.getAvailableTimes(dt=cycle)
        has_prof = cycle in times

        if has_prof:
            stns = self.getAvailableAtTime(dt=cycle)
            if point['icao'] != '':
                has_prof = point['icao'] in [ s['icao'] for s in stns ]
            elif point['iata'] != '':
                has_prof = point['iata'] in [ s['iata'] for s in stns ]
            else:
                has_prof = point['synop'] in [ s['synop'] for s in stns ]
        return has_prof

    def getPoints(self):
        points = self._points
        return points

    def getFields(self):
        return self._csv_fields

    def isAvailable(self):
        return self._is_available


class DataSource(object):
    def __init__(self, config):
        self._name = config.get('name')
        self._ensemble = config.get('ensemble').lower() == "true"
        self._observed = config.get('observed').lower() == "true"
        self._outlets = dict( (c.get('name'), Outlet(self._name, c)) for c in config )

    def _get(self, name, outlet_num=None, flatten=True, **kwargs):
        prop = None
        if outlet_num is None:
            prop = []
            for o in self._outlets.values():
                func = getattr(o, name)
                prop.append(func(**kwargs))

            if flatten:
                prop = [ p for plist in prop for p in plist ]
                prop = list(set(prop))
                prop = sorted(prop)
        else:
            func = getattr(list(self._outlets.values())[outlet_num], name)
            prop = func()
        return prop

    def _getOutletWithProfile(self, stn, cycle_dt, outlet_num=0):
        logging.debug("_getOutletWithProfile: " + str(stn) + ' ' + str(cycle_dt))
        use_outlets = [ out for out, cfg in self._outlets.items() if cfg.hasProfile(stn, cycle_dt) ]
        try:
            outlet = use_outlets[outlet_num]
        except IndexError:
            raise DataSourceError()
        return outlet

    def updateTimeSpan(self):
        outlets = [ self._outlets[key].getTimeSpan() for key in self._outlets.keys() ]
        return outlets

    def getForecastHours(self, outlet_num=None, flatten=True):
        times = self._get('getForecastHours', outlet_num=outlet_num, flatten=flatten)
        return times

    def getDailyCycles(self, outlet_num=None, flatten=True):
        cycles = self._get('getCycles', outlet_num=outlet_num, flatten=flatten)
        return cycles

    def getDelays(self, outlet_num=None):
        delays = self._get('getDelay', outlet, flatten=False)
        return delays

    def getArchiveLens(self, outlet_num=None):
        lens = self._get('getArchiveLen', outlet_num=outlet_num, flatten=False)
        return lens

    def getMostRecentCycle(self, outlet_num=None):
        cycles = self._get('getMostRecentCycle', outlet_num=outlet_num, flatten=False)
        return max(cycles)

    def getAvailableTimes(self, filename=None, outlet_num=None, max_cycles=100, dt=None):
        logging.debug("data_source.getAvailableTimes() filename="+str(filename)+' outlet_num=' +str(outlet_num) + ' dt=' + str(dt))
        cycles = self._get('getAvailableTimes', outlet_num=outlet_num, filename=filename, max_cycles=max_cycles, dt=dt)
        return cycles[-max_cycles:]

    def getAvailableAtTime(self, dt, outlet_num=None):
        points = self._get('getAvailableAtTime', outlet_num=outlet_num, flatten=False, dt=dt)

        flatten_pts = []
        flatten_coords = []
        for pt_list in points:
            for pt in pt_list:
                if (pt['lat'], pt['lon']) not in flatten_coords:
                    flatten_coords.append((pt['lat'], pt['lon']))
                    flatten_pts.append(pt)
        return flatten_pts

    def getDecoder(self, stn, cycle_dt, outlet_num=0):
        outlet = self._getOutletWithProfile(stn, cycle_dt, outlet_num=outlet_num)
        decoder = self._outlets[outlet].getDecoder()
        return decoder

    def getURL(self, stn, cycle_dt, outlet_num=0, outlet=None):
        if outlet is None:
            outlet = self._getOutletWithProfile(stn, cycle_dt, outlet_num=outlet_num)
        url_base = self._outlets[outlet].getURL()
        logging.debug("URL: " + url_base)

        # JTS - Format the URL differently for NUCAPS.
        if self._name == 'NUCAPS CONUS NOAA-20' \
            or self._name == 'NUCAPS CONUS Aqua' \
            or self._name == 'NUCAPS CONUS MetOp-A' \
            or self._name == 'NUCAPS CONUS MetOp-B' \
            or self._name == 'NUCAPS CONUS MetOp-C' \
            or self._name == 'NUCAPS Caribbean NOAA-20' \
            or self._name == 'NUCAPS Alaska NOAA-20':
            fmt = {
                'srcid':quote(stn['srcid']),
                'cycle':"%02d%02d" % (cycle_dt.hour, cycle_dt.minute),
                'date':cycle_dt.strftime("%y%m%d"),
                'year':cycle_dt.strftime("%Y"),
                'month':cycle_dt.strftime("%m"),
                'day':cycle_dt.strftime("%d")
            }
        else:
            fmt = {
                'srcid':quote(stn['srcid']),
                'cycle':"%02d" % cycle_dt.hour,
                'date':cycle_dt.strftime("%y%m%d"),
                'year':cycle_dt.strftime("%Y"),
                'month':cycle_dt.strftime("%m"),
                'day':cycle_dt.strftime("%d")
            }
        url = url_base.format(**fmt)
        logging.debug("Formatted URL: " + url)
        return url

    def getDecoderAndURL(self, stn, cycle_dt, outlet_num=0):
        logging.debug("Getting the decoder and the URL to the data ...")
        outlet = self._getOutletWithProfile(stn, cycle_dt, outlet_num=outlet_num)
        decoder = self._outlets[outlet].getDecoder()
        url = self.getURL(stn, cycle_dt, outlet_num, outlet=outlet)
        return decoder, url

    def getURLList(self, outlet_num=None):
        return self._get('getURL', outlet_num=outlet_num, flatten=False)

    def getName(self):
        return self._name

    def isEnsemble(self):
        return self._ensemble

    def isObserved(self):
        return self._observed

    def getPoint(self, stn):
        for outlet in self._outlets.items():
            for pnt in outlet[1].getPoints():
                if stn in pnt['icao'] or stn in pnt['iata'] or stn in pnt['srcid']:
                    return pnt

if __name__ == "__main__":
    ds = loadDataSources('./')
    ds = dict( (n, ds[n]) for n in ['Observed', 'GFS', 'HRRR', 'NAM'] )
    print(ds['GFS'].updateTimeSpan())
    print(ds['HRRR'].updateTimeSpan())
#    for n, d in ds.items():
#       print n, d.getMostRecentCycle()
        #times = d.getAvailableTimes()
        #for t in times:
        #    print(n, t, [ s['srcid'] for s in d.getAvailableAtTime(t) ])



''' Frequently used functions '''
import numpy as np
import numpy.ma as ma
import numpy.testing as npt
from sharppy.sharptab.constants import MISSING, TOL
from sharppy.sharptab import utils as utils


# vec2comp Tests
def test_vec2comp_single():
    input_wdir = 225
    input_wspd = 7.0710678118654755
    correct_u = 5
    correct_v = 5
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd)
    npt.assert_almost_equal(returned_u, correct_u)
    npt.assert_almost_equal(returned_v, correct_v)

def test_vec2comp_array_like():
    input_wdir = [0, 45, 90, 135, 180, 225, 270, 315, 360]
    input_wspd = [5, 10, 15, 20, 25, 30, 35, 40, 45]
    correct_u = [0, -7.0710678118654746, -15, -14.142135623730951, 0,
        21.213203435596423, 35, 28.284271247461909, 0]
    correct_v = [-5, -7.0710678118654746, 0, 14.142135623730951, 25,
        21.213203435596423, 0, -28.284271247461909, -45]
    correct_u = np.asanyarray(correct_u).astype(np.float64)
    correct_v = np.asanyarray(correct_v).astype(np.float64)
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd)
    npt.assert_almost_equal(returned_u, correct_u)
    npt.assert_almost_equal(returned_v, correct_v)

def test_vec2comp_zeros():
    input_wdir = [0, 90, 180, 270, 360]
    input_wspd = [10, 20, 30, 40, 50]
    correct_u = [0, -20, 0, 40, 0]
    correct_v = [-10, 0, 30, 0, -50]
    correct_u = np.asanyarray(correct_u).astype(np.float64)
    correct_v = np.asanyarray(correct_v).astype(np.float64)
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd)
    npt.assert_equal(returned_u, correct_u)
    npt.assert_equal(returned_v, correct_v)

def test_vec2comp_default_missing_val_single():
    input_wdir = MISSING
    input_wspd = 30
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd)
    npt.assert_equal(type(returned_u), type(ma.masked))
    npt.assert_equal(type(returned_v), type(ma.masked))

def test_vec2comp_default_missing_val_array():
    input_wdir = [0, 90, 180, MISSING]
    input_wspd = [MISSING, 10, 20, 30]
    correct_u = [MISSING, -10, 0, MISSING]
    correct_v= [MISSING, 0, 20, MISSING]
    correct_u = ma.asanyarray(correct_u).astype(np.float64)
    correct_v = ma.asanyarray(correct_v).astype(np.float64)
    correct_u[correct_u == MISSING] = ma.masked
    correct_v[correct_v == MISSING] = ma.masked
    correct_u[correct_v.mask] = ma.masked
    correct_v[correct_u.mask] = ma.masked
    correct_u.set_fill_value(MISSING)
    correct_v.set_fill_value(MISSING)
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd)
    npt.assert_almost_equal(returned_u, correct_u)
    npt.assert_almost_equal(returned_v, correct_v)

def test_vec2comp_user_missing_val_single():
    missing = 50
    input_wdir = missing
    input_wspd = 30
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd, missing)
    npt.assert_equal(type(returned_u), type(ma.masked))
    npt.assert_equal(type(returned_v), type(ma.masked))

def test_vec2comp_user_missing_val_array():
    missing = 50
    input_wdir = [0, 90, 180, missing]
    input_wspd = [missing, 10, 20, 30]
    correct_u = [missing, -10, 0, missing]
    correct_v= [missing, 0, 20, missing]
    correct_u = ma.asanyarray(correct_u).astype(np.float64)
    correct_v = ma.asanyarray(correct_v).astype(np.float64)
    correct_u[correct_u == missing] = ma.masked
    correct_v[correct_v == missing] = ma.masked
    correct_u[correct_v.mask] = ma.masked
    correct_v[correct_u.mask] = ma.masked
    correct_u.set_fill_value(missing)
    correct_v.set_fill_value(missing)
    returned_u, returned_v = utils.vec2comp(input_wdir, input_wspd, missing)
    npt.assert_almost_equal(returned_u, correct_u)
    npt.assert_almost_equal(returned_v, correct_v)


# comp2vec Tests
def test_comp2vec_single():
    input_u = 5
    input_v = 5
    correct_wdir = 225
    correct_wspd = 7.0710678118654755
    returned_wdir, returned_wspd = utils.comp2vec(input_u, input_v)
    npt.assert_almost_equal(returned_wdir, correct_wdir)
    npt.assert_almost_equal(returned_wspd, correct_wspd)

def test_comp2vec_array():
    input_u = [0, -7.0710678118654746, -15, -14.142135623730951, 0,
        21.213203435596423, 35, 28.284271247461909, 0]
    input_v = [-5, -7.0710678118654746, 0, 14.142135623730951, 25,
        21.213203435596423, 0, -28.284271247461909, -45]
    correct_wdir = [0, 45, 90, 135, 180, 225, 270, 315, 0]
    correct_wspd = [5, 10, 15, 20, 25, 30, 35, 40, 45]
    correct_wdir = np.asanyarray(correct_wdir).astype(np.float64)
    correct_wspd = np.asanyarray(correct_wspd).astype(np.float64)
    returned_wdir, returned_wspd = utils.comp2vec(input_u, input_v)
    npt.assert_almost_equal(correct_wdir, returned_wdir)
    npt.assert_almost_equal(correct_wspd, returned_wspd)

def test_comp2vec_zeros():
    input_u = [0, -20, 0, 40, 0]
    input_v = [-10, 0, 30, 0, -50]
    correct_wdir = [0, 90, 180, 270, 0]
    correct_wspd = [10, 20, 30, 40, 50]
    correct_wdir = np.asanyarray(correct_wdir).astype(np.float64)
    correct_wspd = np.asanyarray(correct_wspd).astype(np.float64)
    returned_wdir, returned_wspd = utils.comp2vec(input_u, input_v)
    npt.assert_equal(returned_wdir, correct_wdir)
    npt.assert_equal(returned_wspd, correct_wspd)

def test_comp2vec_default_missing_val_single():
    input_u = MISSING
    input_v = 30
    returned_wdir, returned_wspd = utils.comp2vec(input_u, input_v)
    npt.assert_equal(type(returned_wdir), type(ma.masked))
    npt.assert_equal(type(returned_wspd), type(ma.masked))

def test_comp2vec_default_missing_val_array():
    input_u = [MISSING, -10, 0, MISSING]
    input_v= [MISSING, 0, 20, MISSING]
    correct_wdir = [0, 90, 180, MISSING]
    correct_wspd = [MISSING, 10, 20, 30]
    correct_wdir = ma.asanyarray(correct_wdir).astype(np.float64)
    correct_wspd = ma.asanyarray(correct_wspd).astype(np.float64)
    correct_wdir[correct_wdir == MISSING] = ma.masked
    correct_wspd[correct_wspd == MISSING] = ma.masked
    correct_wdir[correct_wspd.mask] = ma.masked
    correct_wspd[correct_wdir.mask] = ma.masked
    correct_wdir.set_fill_value(MISSING)
    correct_wspd.set_fill_value(MISSING)
    returned_wdir, returned_wspd = utils.comp2vec(input_u, input_v)
    npt.assert_almost_equal(returned_wdir, correct_wdir)
    npt.assert_almost_equal(returned_wspd, correct_wspd)

def test_comp2vec_user_missing_val_single():
    missing = 50
    input_u = missing
    input_v = 30
    returned_wdir, returned_wspd = utils.vec2comp(input_u, input_v, missing)
    npt.assert_equal(type(returned_wdir), type(ma.masked))
    npt.assert_equal(type(returned_wspd), type(ma.masked))

def test_comp2vec_user_missing_val_array():
    missing = 50
    input_u = [missing, -10, 0, missing]
    input_v= [missing, 0, 20, missing]
    correct_wdir = [0, 90, 180, missing]
    correct_wspd = [missing, 10, 20, 30]
    correct_wdir = ma.asanyarray(correct_wdir).astype(np.float64)
    correct_wspd = ma.asanyarray(correct_wspd).astype(np.float64)
    correct_wdir[correct_wdir == missing] = ma.masked
    correct_wspd[correct_wspd == missing] = ma.masked
    correct_wdir[correct_wspd.mask] = ma.masked
    correct_wspd[correct_wdir.mask] = ma.masked
    correct_wdir.set_fill_value(missing)
    correct_wspd.set_fill_value(missing)
    returned_wdir, returned_wspd = utils.comp2vec(input_u, input_v)
    npt.assert_almost_equal(returned_wdir, correct_wdir)
    npt.assert_almost_equal(returned_wspd, correct_wspd)


# mag Tests
def test_mag_single():
    input_u = 5
    input_v = 5
    correct_answer = np.sqrt(input_u**2 + input_v**2)
    returned_answer = utils.mag(input_u, input_v)
    npt.assert_almost_equal(returned_answer, correct_answer)

def test_mag_zero():
    input_u = 0
    input_v = 0
    correct_answer = 0
    returned_answer = utils.mag(input_u, input_v)
    npt.assert_almost_equal(returned_answer, correct_answer)

def test_mag_array():
    rt2 = np.sqrt(2)
    input_u = [5, 10, 15]
    input_v = [5, 10, 15]
    correct_answer = [5*rt2, 10*rt2, 15*rt2]
    correct_answer = ma.asanyarray(correct_answer)
    returned_answer = utils.mag(input_u, input_v)
    npt.assert_almost_equal(returned_answer, correct_answer)

def test_mag_default_missing_single():
    input_u = MISSING
    input_v = 10
    correct_answer = ma.masked
    returned_answer = utils.mag(input_u, input_v)
    npt.assert_equal(type(returned_answer), type(correct_answer))

def test_mag_default_missing_array():
    rt2 = np.sqrt(2)
    input_u = [MISSING, 10, 20, 30, 40]
    input_v = [0, 10, 20, 30, MISSING]
    correct_answer = [MISSING, 10*rt2, 20*rt2, 30*rt2, MISSING]
    correct_answer = ma.asanyarray(correct_answer).astype(np.float64)
    correct_answer[correct_answer == MISSING] = ma.masked
    returned_answer = utils.mag(input_u, input_v)
    npt.assert_almost_equal(returned_answer, correct_answer)

def test_mag_user_missing_single():
    missing = 50
    input_u = missing
    input_v = 10
    correct_answer = ma.masked
    returned_answer = utils.mag(input_u, input_v, missing)
    npt.assert_equal(type(returned_answer), type(correct_answer))

def test_mag_user_missing_array():
    missing = 50
    rt2 = np.sqrt(2)
    input_u = [missing, 10, 20, 30, 40]
    input_v = [0, 10, 20, 30, missing]
    correct_answer = [missing, 10*rt2, 20*rt2, 30*rt2, missing]
    correct_answer = ma.asanyarray(correct_answer).astype(np.float64)
    correct_answer[correct_answer == missing] = ma.masked
    returned_answer = utils.mag(input_u, input_v, missing)
    npt.assert_almost_equal(returned_answer, correct_answer)



import numpy as np
import sharppy
from qtpy import QtGui, QtCore, QtWidgets
from qtpy.QtWidgets import *
from datetime import datetime
import sys, os
import re

class Mapper(object):
    data_dir = os.path.join(os.path.dirname(sharppy.__file__), 'databases', 'shapefiles')
    min_lat = {'npstere':0., 'merc':-35., 'spstere':-90.}
    max_lat = {'npstere':90., 'merc':35., 'spstere':0.}

    def __init__(self, lambda_0, phi_0, proj='npstere'):
        self.proj = proj
        self.lambda_0 = lambda_0
        self.phi_0 = phi_0
        if proj == 'spstere':
            self.phi_0 = -np.abs(self.phi_0)

        self.m = 6.6667e-7
        self.rad_earth = 6.371e8
        self._bnds = {}

    def getLambda0(self):
        return self.lambda_0

    def setLambda0(self, lambda_0):
        self.lambda_0 = lambda_0

    def getPhi0(self):
        return self.phi_0

    def setProjection(self, proj):
        if proj not in ['npstere', 'spstere', 'merc']:
            raise ValueError("Projection must be one of 'npstere', 'spstere', or 'merc'; got '%s'." % proj)
        self.proj = proj

        if proj == 'spstere':
            self.phi_0 = -np.abs(self.phi_0)
        elif proj == 'npstere':
            self.phi_0 = np.abs(self.phi_0)

    def getProjection(self):
        return self.proj

    def getCoordPaths(self):
        path = QtGui.QPainterPath()
        lb_lat, ub_lat = Mapper.min_lat[self.proj], Mapper.max_lat[self.proj]

        if self.proj == 'npstere':
            for lon in range(0, 360, 20):
                lats = np.linspace(lb_lat, ub_lat, 2)
                lx, ly = self(lats, lon)

                path.moveTo(lx[0], ly[0])
                for x, y in zip(lx[1:], ly[1:]):
                    path.lineTo(x, y)

            for lat in range(int(lb_lat), int(ub_lat), 15):
                lons = np.arange(self.getLambda0(), self.getLambda0() + 360, 90)
                rx, ry = self(lat, lons)
                x_min, x_max = rx.min(), rx.max()
                y_min, y_max = ry.min(), ry.max()
                path.addEllipse(x_min, y_min, x_max - x_min, y_max - y_min)

        elif self.proj == 'merc':
            for lon in range(-180, 180 + 20, 20):
                lats = np.linspace(lb_lat, ub_lat, 2)
                lx, ly = self(lats, lon)

                path.moveTo(lx[0], ly[0])
                for x, y in zip(lx[1:], ly[1:]):
                    path.lineTo(x, y)

            lat_spc = 10
            rnd_lat_lb = np.ceil(lb_lat / lat_spc) * lat_spc
            rnd_lat_ub = np.floor(ub_lat / lat_spc) * lat_spc
            lat_lines = list(range(int(rnd_lat_lb), int(rnd_lat_ub + lat_spc), lat_spc))
            if rnd_lat_lb != lb_lat:
                lat_lines = [int(lb_lat)] + lat_lines
            if rnd_lat_ub != ub_lat:
                lat_lines = lat_lines + [int(ub_lat)]

            for lat in lat_lines:
                lons = np.linspace(-180, 180, 2)
                lx, ly = self(lat, lons)

                path.moveTo(lx[0], ly[0])
                for x, y in zip(lx[1:], ly[1:]):
                    path.lineTo(x, y)

        elif self.proj == 'spstere':
            for lon in range(0, 360, 20):
                lats = np.linspace(lb_lat, ub_lat, 2)
                lx, ly = self(lats, lon)

                path.moveTo(lx[0], ly[0])
                for x, y in zip(lx[1:], ly[1:]):
                    path.lineTo(x, y)

            for lat in range(int(ub_lat), int(lb_lat), -15):
                lons = np.arange(self.getLambda0(), self.getLambda0() + 360, 90)
                rx, ry = self(lat, lons)
                x_min, x_max = rx.min(), rx.max()
                y_min, y_max = ry.min(), ry.max()
                path.addEllipse(x_min, y_min, x_max - x_min, y_max - y_min)

        return path

    def getLatBounds(self):
        return Mapper.min_lat[self.proj], Mapper.max_lat[self.proj]

    def __call__(self, coord1, coord2, inverse=False):
        if inverse:
            if self.proj in ['npstere', 'spstere']:
                return self._xytoll_stere(coord1, coord2, self.lambda_0, self.phi_0, self.m, self.rad_earth)
            elif self.proj in ['merc']:
                return self._xytoll_merc(coord1, coord2, self.lambda_0, self.m, self.rad_earth)
        else:
            if self.proj in ['npstere', 'spstere']:
                return self._lltoxy_stere(coord1, coord2, self.lambda_0, self.phi_0, self.m, self.rad_earth)
            elif self.proj in ['merc']:
                return self._lltoxy_merc(coord1, coord2, self.lambda_0, self.m, self.rad_earth)

    # Functions to perform the map transformation to North Pole Stereographic
    # Equations from the SoM OBAN 2014 class
    # Functions adapted for either hemisphere and inverse transformations added by Tim Supinie, April 2015
    def _get_sigma(self, phi_0, lats, south_hemis=False):
        sign = -1 if south_hemis else 1
        sigma = (1. + np.sin(np.radians(sign * phi_0))) / (1. + np.sin(np.radians(sign * lats)))
        return sigma

    def _get_shifted_lon(self, lambda_0, lons, south_hemis=False):
        sign = -1 if south_hemis else 1
        return sign * (lambda_0 - lons)

    def _lltoxy_stere(self, lats, lons, lambda_0, phi_0, m, rad_earth):
        sigma = self._get_sigma(phi_0, lats, south_hemis=(phi_0 < 0))
        lambdas = np.radians(self._get_shifted_lon(lambda_0 + 90, lons, south_hemis=(phi_0 < 0)))
        x = m * sigma * rad_earth * np.cos(np.radians(lats)) * np.cos(lambdas)
        y = m * sigma * rad_earth * np.cos(np.radians(lats)) * np.sin(lambdas)
        return x, y

    def _xytoll_stere(self, xs, ys, lambda_0, phi_0, m, rad_earth):
        sign = -1 if (phi_0 < 0) else 1

        lon = (lambda_0 + 90 - sign * np.degrees(np.arctan2(ys, xs)))
        lat = sign * np.degrees(2 * np.arctan(rad_earth * m * (1 + sign * np.sin(np.radians(phi_0))) / np.hypot(xs, ys)) - np.pi / 2)

        if lon < -180: lon += 360
        elif lon > 180: lon -= 360
        return lat, lon

    # Function to perform map transformation to and from Mercator projection
    def _lltoxy_merc(self, lats, lons, lambda_0, m, rad_earth):
        x = m * rad_earth * (np.radians(lons) - np.radians(lambda_0))
        y = -m * rad_earth * np.log(np.tan(np.pi / 4 + np.radians(lats) / 2))

        if type(x) in [ np.ndarray ] or type(y) in [ np.ndarray ]:
            if type(x) not in [ np.ndarray ]:
                x = x * np.ones(y.shape)
            if type(y) not in [ np.ndarray ]:
                y = y * np.ones(x.shape)
        return x, y

    def _xytoll_merc(self, xs, ys, lambda_0, m, rad_earth):
        lon = np.degrees(np.radians(lambda_0) + xs / (m * rad_earth))
        lat = -np.degrees(2 * np.arctan(np.exp(ys / (m * rad_earth))) - np.pi / 2)
        return lat, lon

    def _loadDat(self, name, res):
        """
        Code shamelessly lifted from Basemap's data file parser by Jeff Whitaker.
        http://matplotlib.org/basemap/
        """
        def segmentPath(b, lb_lat, ub_lat):
            paths = []
            if b[:, 1].max() <= lb_lat or b[:, 1].min() >= ub_lat:
                return paths

            idxs = np.where((b[:, 1] >= lb_lat) & (b[:, 1] <= ub_lat))[0]
            if len(idxs) < 2:
                return paths

            segs = (np.diff(idxs) == 1)
            try:
                breaks = np.where(segs == 0)[0] + 1
            except IndexError:
                breaks = []
            breaks = [ 0 ] + list(breaks) + [ -1 ]

            for idx in range(len(breaks) - 1):
                if breaks[idx + 1] == -1:
                    seg_idxs = idxs[breaks[idx]:]
                else:
                    seg_idxs = idxs[breaks[idx]:breaks[idx + 1]]

                if len(seg_idxs) >= 2:
                    paths.append(b[seg_idxs, ::-1])

            return paths

        bdatfile = open(os.path.join(Mapper.data_dir, name + '_' + res + '.dat'), 'rb')
        bdatmetafile = open(os.path.join(Mapper.data_dir, name + 'meta_' + res + '.dat'), 'r')

        projs = ['npstere', 'merc', 'spstere']
        paths = dict( (p, []) for p in projs )

#       old_proj = self.proj
        for line in bdatmetafile:
            lats, lons = [], []
            linesplit = line.split()
            area = float(linesplit[1])
            south = float(linesplit[3])
            north = float(linesplit[4])
            if area < 0:
                area = 1e30

            if area > 1500.:
                typ = int(linesplit[0])
                npts = int(linesplit[2])
                offsetbytes = int(linesplit[5])
                bytecount = int(linesplit[6])
                bdatfile.seek(offsetbytes,0)
                # read in binary string convert into an npts by 2
                # numpy array (first column is lons, second is lats).
                polystring = bdatfile.read(bytecount)
                # binary data is little endian.
                #b = np.array(np.fromstring(polystring,dtype='<f4'),'f8')
                b = np.array(np.frombuffer(polystring,dtype='<f4'),'f8')
                b.shape = (npts, 2)

                if np.any(b[:, 0] > 180):
                    b[:, 0] -= 360

                for proj in projs:
                    lb_lat, ub_lat = Mapper.min_lat[proj], Mapper.max_lat[proj]
                    path = segmentPath(b, lb_lat, ub_lat)
                    paths[proj].extend(path)

        return paths


    def getBoundary(self, name):
        if name == 'coastlines':
            name = 'gshhs'

        if name == 'states':
            res = 'h'
        elif name == 'uscounties':
            res = 'f'
        else:
            res = 'i'

        if name not in self._bnds:
            self._bnds[name] = self._loadDat(name, res)

        paths = []
        for bnd in self._bnds[name][self.proj]:
            path = QtGui.QPainterPath()

            path_lats, path_lons = list(zip(*bnd))
            path_x, path_y = self(np.array(path_lats), np.array(path_lons))

            path.moveTo(path_x[0], path_y[0])
            for px, py in zip(path_x[1:], path_y[1:]):
                path.lineTo(px, py)

            paths.append(path)
        return paths

class MapWidget(QWidget):
    clicked = QtCore.Signal(dict)

    def __init__(self, data_source, init_time, async_object, **kwargs):
        config = kwargs.get('cfg', None)
        del kwargs['cfg']

        super(MapWidget, self).__init__(**kwargs)

        self.trans_x, self.trans_y = 0., 0.
        self.center_x, self.center_y = 0., 0.
        self.init_drag_x, self.init_drag_y = None, None
        self.dragging = False
        self.map_rot = 0.0
        self.setMouseTracking(True)

        self.has_internet = True

        self.init_scale = 0.6
        if config is None or not ('map', 'proj') in config:
            self.scale = self.init_scale
            self.map_center_x, self.map_center_y = 0., 0.
            std_lon = -97.5
            proj = 'npstere'
            init_from_config = False
        else:
            proj = config['map', 'proj']
            std_lon = float(config['map', 'std_lon'])
            self.scale = float(config['map', 'scale'])
            self.map_center_x = float(config['map', 'center_x'])
            self.map_center_y = float(config['map', 'center_y'])
            init_from_config = True

        self.mapper = Mapper(std_lon, 60., proj=proj)

        self.stn_lats = np.array([])
        self.stn_lons = np.array([])
        self.stn_ids = []
        self.stn_names = []
        self.stn_qc_flags = [] # JTS

        self.default_width, self.default_height = self.width(), self.height()
        self.setMinimumSize(self.width(), self.height())

        self.clicked_stn = None
        self.stn_readout = QtWidgets.QLabel(parent=self)
        self.stn_readout.setStyleSheet("QLabel { background-color:#000000; border-width: 0px; font-size: 16px; color: #FFFFFF; }")
        self.stn_readout.setText("")
        self.stn_readout.show()
        self.stn_readout.move(self.width(), self.height())

        self.load_readout = QtWidgets.QLabel(parent=self)
        self.load_readout.setStyleSheet("QLabel { background-color:#000000; border-width: 0px; font-size: 18px; color: #FFFFFF; }")
        self.load_readout.setText("Loading ...")
        self.load_readout.setFixedWidth(100)
        self.load_readout.show()
        self.load_readout.move(self.width(), self.height())

        self.latlon_readout = QtWidgets.QLabel(parent=self)
        self.latlon_readout.setStyleSheet("QLabel { background-color:#000000; border-width: 0px; font-size: 18px; color: #FFFFFF; }")
        self.latlon_readout.setText("")
        self.latlon_readout.setFixedWidth(150)
        self.latlon_readout.show()
        self.latlon_readout.move(10, 10)

        self.no_internet = QtWidgets.QLabel(parent=self)
        self.no_internet.setStyleSheet("QLabel { background-color:#000000; border-width: 0px; font-size: 32px; color: #FFFFFF; }")
        self.no_internet.setText("No Internet Connection")
        self.no_internet.show()
        txt_width = self.no_internet.fontMetrics().width(self.no_internet.text())

        self._async = async_object
        self.no_internet.setFixedWidth(txt_width)
        self.no_internet.move(self.width(), self.height())

        self.setDataSource(data_source, init_time, init=True)
        self.setWindowTitle('SHARPpy')

        if not init_from_config:
            self.resetViewport()
            self.saveProjection(config)

        self.pt_x, self.pt_y = None, None

        self.initMap()
        self.initUI()

    def initUI(self):
        self.center_x, self.center_y = self.width() / 2, self.height() / 2

        self.plotBitMap = QtGui.QPixmap(self.width(), self.height())
        self.backgroundBitMap = self.plotBitMap.copy()
        self.drawMap()

    def initMap(self):
        self._coast_path = self.mapper.getBoundary('coastlines')
        self._country_path = self.mapper.getBoundary('countries')
        self._state_path = self.mapper.getBoundary('states')
        self._county_path = self.mapper.getBoundary('uscounties')

        self._grid_path = self.mapper.getCoordPaths()

    def setDataSource(self, data_source, data_time, init=False):
        self.cur_source = data_source
        self.setCurrentTime(data_time, init=init)

    def setCurrentTime(self, data_time, init=False):

        self.clicked_stn = None
        self.clicked.emit(None)

        self._showLoading()

        def update(points):
            self.points = points[0]

            self.stn_lats = np.array([ p['lat'] for p in self.points ])
            self.stn_lons = np.array([ p['lon'] for p in self.points ])
            self.stn_ids = [ p['srcid'] for p in self.points ]

            # JTS - Extract QC flag (priority header) info for NUCAPS data.
            if self.cur_source.getName().startswith("NUCAPS"):
                self.stn_qc_flags = [ p['priority'] for p in self.points ]

            self.stn_names = []

            # JTS - Text that displays when hovering mouse over profiles on the map.
            for p in self.points:
                if self.cur_source.getName().startswith("NUCAPS"):
                    # Format the latitude string
                    if p['lat'] != "":
                        lat_str = " (%s," % p['lat']
                    else:
                        lat_str = ""

                    # Format the longitude string
                    if p['lon'] != "":
                        lon_str = " %s)" % p['lon']
                    else:
                        lon_str = ""

                    # Format the string that will be displayed to the user.
                    nm_before = p['name']
                    date = nm_before.split('_')[0]
                    time = nm_before.split('_')[1]
                    footprint_number = nm_before.split('_')[2]
                    satID = nm_before.split('_')[3]

                    year = date[0:2]
                    month = date[2:4]
                    day = date[4:6]
                    hour = time[0:2]
                    minute = time[2:4]
                    second = time[4:6]

                    nm_after = f'{month}-{day}-{year}_{hour}:{minute}:{second}Z_{footprint_number}_{satID}'
                    nm_after = nm_after.upper()
                    name = "%s%s%s" % (nm_after, lat_str, lon_str)
                    self.stn_names.append(name)
                else:
                    if p['icao'] != "":
                        id_str = " (%s)" % p['icao']
                    else:
                        id_str = ""

                    if p['state'] != "":
                        pol_str = ", %s" % p['state']
                    elif p['country'] != "":
                        pol_str = ", %s" % p['country']
                    else:
                        pol_str = ""

                    nm = p['name']
                    if id_str == "" and pol_str == "":
                        nm = nm.upper()
                    name = "%s%s%s" % (nm, pol_str, id_str)
                    self.stn_names.append(name)

            self._hideLoading()

            if not init:
                self.drawMap()
                self.update()

        self.current_time = data_time
        getPoints = lambda: self.cur_source.getAvailableAtTime(self.current_time)

        if init:
            points = getPoints()
            update([ points ])
        else:
            self._async.post(getPoints, update)

    def setProjection(self, proj):
        old_proj = self.mapper.getProjection()
        self.mapper.setProjection(proj)
        self.resetViewport()
        self._showLoading()

        def update(args):
            self.resetViewport()
            self.drawMap()
            self._hideLoading()
            self.update()
            return

        self._async.post(self.initMap, update)

    def resetViewport(self, ctr_lat=None, ctr_lon=None):
        self.map_center_x = self.width() / 2
        if ctr_lat is not None and ctr_lon is not None:
            center_x, center_y = self.mapper(ctr_lat, ctr_lon)
            self.map_center_y = -center_y + self.height() / 2
            self.map_center_y = self.center_y - (self.center_y - self.map_center_y) / self.scale
        else:
            self.scale = self.init_scale
            self.map_rot = 0.
            proj = self.mapper.getProjection()
            if proj == 'npstere':
                self.map_center_y = -13 * self.height() / 10 + self.height() / 2
            elif proj == 'merc':
                self.map_center_y = self.height() / 2
            elif proj == 'spstere':
                self.map_center_y = 13 * self.height() / 10 + self.height() / 2

    def drawMap(self):
        def mercRotate(qp, center_x, center_y, angle):
            center_lat, center_lon = self.mapper(center_x - self.width() / 2, center_y - self.height() / 2, inverse=True)
            center_lon -= angle
            new_center_x, new_center_y = self.mapper(center_lat, center_lon)
            new_center_x += self.width() / 2
            new_center_y += self.height() / 2
            qp.translate((new_center_x - center_x) / self.scale, (new_center_y - center_y) / self.scale)
            return new_center_x, new_center_y

        qp = QtGui.QPainter()
        qp.begin(self.plotBitMap)

        self.plotBitMap.fill(QtCore.Qt.black)

        map_center_x = self.map_center_x + self.trans_x
        map_center_y = self.map_center_y + self.trans_y

        qp.translate(map_center_x, map_center_y)
        proj = self.mapper.getProjection()

        if proj == 'npstere':
            qp.rotate(self.map_rot)
        elif proj == 'merc':
            new_center_x, new_center_y = mercRotate(qp, map_center_x, map_center_y, self.map_rot)
        elif proj == 'spstere':
            qp.rotate(-self.map_rot)

        qp.scale(1. / self.scale, 1. / self.scale)

        self.drawPolitical(qp)
        self.backgroundBitMap = self.plotBitMap.copy()
        self.drawStations(qp)

        qp.end()

    def drawPolitical(self, qp):
        self.transform = qp.transform()
        window_rect = QtCore.QRect(0, 0, self.width(), self.height())

        qp.setPen(QtGui.QPen(QtGui.QColor('#333333'), self.scale)) #, self.scale, QtCore.Qt.DashLine
        qp.drawPath(self._grid_path)

        # Modify the scale thresholds according to the ratio of the area of the plot to the default area
        default_area = self.default_width * self.default_height
        actual_area = self.width() * self.height()
        scaled_area = np.sqrt(default_area / float(actual_area))

        if self.scale < 0.15 * scaled_area:

            max_comp = 102
            full_scale = 0.10 * scaled_area
            zero_scale = 0.15 * scaled_area
            comp = max_comp * min(max((zero_scale - self.scale) / (zero_scale - full_scale), 0), 1)
            color = '#' + ("{0:02x}".format(int(round(comp)))) * 3

            qp.setPen(QtGui.QPen(QtGui.QColor(color), self.scale))
            for cp in self._county_path:
                if self.transform.mapRect(cp.boundingRect()).intersects(window_rect):
                    qp.drawPath(cp)

        qp.setPen(QtGui.QPen(QtGui.QColor('#999999'), self.scale))
        for sp in self._state_path:
            if self.transform.mapRect(sp.boundingRect()).intersects(window_rect):
                qp.drawPath(sp)

        qp.setPen(QtGui.QPen(QtCore.Qt.white, self.scale))
        for cp in self._coast_path:
            if self.transform.mapRect(cp.boundingRect()).intersects(window_rect):
                qp.drawPath(cp)

        for cp in self._country_path:
            if self.transform.mapRect(cp.boundingRect()).intersects(window_rect):
                qp.drawPath(cp)

    def drawStations(self, qp):
        stn_xs, stn_ys = self.mapper(self.stn_lats, self.stn_lons + self.map_rot)
        lb_lat, ub_lat = self.mapper.getLatBounds()
        size = 3 * self.scale

        # JTS - Apply QC coloring to NUCAPS.
        if self.cur_source.getName().startswith("NUCAPS"):
            unselected_color_nucaps = QtCore.Qt.red
            selected_color_nucaps = QtCore.Qt.green

            window_rect = QtCore.QRect(0, 0, self.width(), self.height())
            clicked_x, clicked_y, clicked_lat, clicked_id = None, None, None, None
            for stn_x, stn_y, stn_lat, stn_id, stn_qc_flag in zip(stn_xs, stn_ys, self.stn_lats, self.stn_ids, self.stn_qc_flags):
                if self.clicked_stn == stn_id:
                    clicked_x = stn_x
                    clicked_y = stn_y
                    clicked_lat = stn_lat
                    clicked_id = stn_id
                else:
                    if stn_qc_flag == 'green':
                        unselected_color_nucaps = QtCore.Qt.darkGreen
                    elif  stn_qc_flag == 'yellow':
                        unselected_color_nucaps = QtCore.Qt.yellow
                    elif  stn_qc_flag == 'red':
                        unselected_color_nucaps = QtCore.Qt.red
                    else:
                        unselected_color_nucaps = QtCore.Qt.white

                    if lb_lat <= stn_lat and stn_lat <= ub_lat and window_rect.contains(*self.transform.map(stn_x, stn_y)):
                        qp.setPen(QtGui.QPen(unselected_color_nucaps, self.scale))
                        qp.setBrush(QtGui.QBrush(unselected_color_nucaps))
                        qp.drawEllipse(QtCore.QPointF(stn_x, stn_y), size, size)

            if clicked_lat is not None and lb_lat <= clicked_lat and clicked_lat <= ub_lat and window_rect.contains(*self.transform.map(clicked_x, clicked_y)):
                qp.setPen(QtGui.QPen(selected_color_nucaps, self.scale))
                qp.setBrush(QtGui.QBrush(selected_color_nucaps))
                qp.drawEllipse(QtCore.QPointF(clicked_x, clicked_y), size, size)

        else: # JTS - Non-NUCAPS data will not have the QC coloring applied.
            unselected_color = QtCore.Qt.red
            selected_color = QtCore.Qt.green

            window_rect = QtCore.QRect(0, 0, self.width(), self.height())
            clicked_x, clicked_y, clicked_lat, clicked_id = None, None, None, None

            color = unselected_color
            for stn_x, stn_y, stn_lat, stn_id in zip(stn_xs, stn_ys, self.stn_lats, self.stn_ids):
                if self.clicked_stn == stn_id:
                    clicked_x = stn_x
                    clicked_y = stn_y
                    clicked_lat = stn_lat
                    clicked_id = stn_id
                else:
                    if lb_lat <= stn_lat and stn_lat <= ub_lat and window_rect.contains(*self.transform.map(stn_x, stn_y)):
                        qp.setPen(QtGui.QPen(color, self.scale))
                        qp.setBrush(QtGui.QBrush(color))
                        qp.drawEllipse(QtCore.QPointF(stn_x, stn_y), size, size)

            color = selected_color
            if clicked_lat is not None and lb_lat <= clicked_lat and clicked_lat <= ub_lat and window_rect.contains(*self.transform.map(clicked_x, clicked_y)):
                qp.setPen(QtGui.QPen(color, self.scale))
                qp.setBrush(QtGui.QBrush(color))
                qp.drawEllipse(QtCore.QPointF(clicked_x, clicked_y), size, size)

            if self.cur_source.getName() == "Local WRF-ARW" and self.pt_x != None:
                qp.drawEllipse(QtCore.QPointF(self.pt_x, self.pt_y), size, size)

    def paintEvent(self, e):
        qp = QtGui.QPainter()
        qp.begin(self)
        qp.drawPixmap(0, 0, self.plotBitMap)
        qp.end()

    def resizeEvent(self, e):
        old_size = e.oldSize()
        new_size = e.size()

        if old_size.width() == -1 and old_size.height() == -1:
            old_size = self.size()

        self.map_center_x += (new_size.width() - old_size.width()) / 2.
        self.map_center_y += (new_size.height() - old_size.height()) / 2.
        self._hideLoading()
        self.hasInternet(self.has_internet)

        self.initUI()

    def keyPressEvent(self, e):
        if e.key() == 61:
            delta = -100
        elif e.key() == 45:
            delta = 100
        scale_fac = 10 ** (delta / 1000.)

        scaled_size = float(min(self.default_width, self.default_height)) / min(self.width(), self.height())
        if self.scale * scale_fac > 2.5 * scaled_size:
            scale_fac = 2.5 * scaled_size / self.scale

        self.scale *= scale_fac
        self.map_center_x = self.center_x - (self.center_x - self.map_center_x) / scale_fac
        self.map_center_y = self.center_y - (self.center_y - self.map_center_y) / scale_fac

        self.drawMap()
        self.update()

    def mousePressEvent(self, e):
        self.init_drag_x, self.init_drag_y = e.x(), e.y()

    def mouseMoveEvent(self, e):
        if self.init_drag_x is not None and self.init_drag_y is not None:
            self.dragging = True
            self.trans_x = e.x() - self.init_drag_x
            self.trans_y = e.y() - self.init_drag_y
            self.drawMap()
            self.update()
        self._checkStations(e)

        trans_inv, is_invertible = self.transform.inverted()
        mouse_x, mouse_y = trans_inv.map(e.x(), e.y())
        lat, lon = self.mapper(mouse_x, mouse_y, inverse=True)
        lon -= self.map_rot

        if lon > 180:
            lon -= 360.
        elif lon <= -180:
            lon += 360.

        self.latlon_readout.setText("%.3f; %.3f" % (lat, lon))

    def mouseReleaseEvent(self, e):
        self.init_drag_x, self.init_drag_y = None, None
        self.map_center_x += self.trans_x
        self.map_center_y += self.trans_y
        self.trans_x, self.trans_y = 0, 0

        if not self.dragging and len(self.stn_lats) > 0:
            lb_lat, ub_lat = self.mapper.getLatBounds()
            idxs = np.array([ idx for idx, slat in enumerate(self.stn_lats) if lb_lat <= slat <= ub_lat ])

            stn_xs, stn_ys = self.mapper(self.stn_lats[idxs], self.stn_lons[idxs] + self.map_rot)
            stn_xs, stn_ys = list(zip(*[ self.transform.map(sx, sy) for sx, sy in zip(stn_xs, stn_ys)  ]))
            stn_xs = np.array(stn_xs)
            stn_ys = np.array(stn_ys)
            dists = np.hypot(stn_xs - e.x(), stn_ys - e.y())
            stn_idx = np.argmin(dists)
            if dists[stn_idx] <= 5:
                self.clicked_stn = self.stn_ids[idxs[stn_idx]]
                self.clicked.emit(self.points[idxs[stn_idx]])

                self.drawMap()
                self.update()
        if not self.dragging and self.cur_source.getName() == "Local WRF-ARW":
            trans_inv, is_invertible = self.transform.inverted()
            self.pt_x, self.pt_y = trans_inv.map(e.x(), e.y())
            lat, lon = self.mapper(self.pt_x, self.pt_y, inverse=True)
            lon -= self.map_rot
            self.clicked.emit((lon, lat))
            self.drawMap()
            self.update()

        self.dragging = False

    def mouseDoubleClickEvent(self, e):
        trans_inv, is_invertible = self.transform.inverted()
        mouse_x, mouse_y = trans_inv.map(e.x(), e.y())
        lat, lon = self.mapper(mouse_x, mouse_y, inverse=True)
        lon -= self.map_rot

        self.map_rot += (lon - self.mapper.getLambda0())

#       if self.map_rot > 180:
#           self.map_rot -= 360
#       if self.map_rot <= -180:
#           self.map_rot += 360

        self.mapper.setLambda0(lon)
        self._showLoading()

        def update(args):
            self.resetViewport(ctr_lat=lat, ctr_lon=lon)
            self.drawMap()
            self._hideLoading()
            self.update()
            return

        update(None)
#       self._async.post(self.initMap, update)

    def wheelEvent(self, e):
        max_speed = 75
        delta = max(min(-e.delta(), max_speed), -max_speed)
        scale_fac = 10 ** (delta / 1000.)

        scaled_size = float(min(self.default_width, self.default_height)) / min(self.width(), self.height())
        if self.scale * scale_fac > 2.5 * scaled_size:
            scale_fac = 2.5 * scaled_size / self.scale

        self.scale *= scale_fac
        self.map_center_x = self.center_x - (self.center_x - self.map_center_x) / scale_fac
        self.map_center_y = self.center_y - (self.center_y - self.map_center_y) / scale_fac

        self.drawMap()
        self._checkStations(e)
        self.update()

    def saveProjection(self, config):
        map_center_x = self.map_center_x + (self.default_width  - self.width() ) / 2.
        map_center_y = self.map_center_y + (self.default_height - self.height()) / 2.

        config['map', 'proj']     = self.mapper.getProjection()
        config['map', 'std_lon']  = self.mapper.getLambda0()
        config['map', 'scale']    = self.scale
        config['map', 'center_x'] = map_center_x
        config['map', 'center_y'] = map_center_y

    def getBackground(self):
        return self.backgroundBitMap

    def hasInternet(self, has_connection):
        self.has_internet = has_connection

        if has_connection:
            self.no_internet.move(self.width(), self.height())
        else:
            met = self.no_internet.fontMetrics()
            txt_width = met.width(self.no_internet.text())
            txt_height = met.height()
            self.no_internet.move((self.width() - txt_width) / 2, (self.height() - txt_height) / 2)

    def _showLoading(self):
        self.load_readout.move(10, self.height() - 25)
        self.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents, True)

    def _hideLoading(self):
        self.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents, False)
        self.load_readout.move(self.width(), self.height())

    def _checkStations(self, e):
#       if len(self.stn_lats) == 0:
#           return

        lb_lat, ub_lat = self.mapper.getLatBounds()
        idxs = np.array([ idx for idx, slat in enumerate(self.stn_lats) if lb_lat <= slat <= ub_lat ])
        if len(idxs) == 0:
            return

        stn_xs, stn_ys = self.mapper(self.stn_lats[idxs], self.stn_lons[idxs] + self.map_rot)
        stn_xs, stn_ys = list(zip(*[ self.transform.map(sx, sy) for sx, sy in zip(stn_xs, stn_ys) ]))
        stn_xs = np.array(stn_xs)
        stn_ys = np.array(stn_ys)
        dists = np.hypot(stn_xs - e.x(), stn_ys - e.y())
        stn_idx = np.argmin(dists)
        if dists[stn_idx] <= 5:
            stn_x, stn_y = stn_xs[stn_idx], stn_ys[stn_idx]
            fm = QtGui.QFontMetrics(QtGui.QFont(self.font().rawName(), 16))

            label_offset = 5
            align = 0
            if stn_x > self.width() / 2:
                sgn_x = -1
                label_x = stn_x - fm.width(self.stn_names[idxs[stn_idx]])
                align |= QtCore.Qt.AlignRight
            else:
                sgn_x = 1
                label_x = stn_x
                align |= QtCore.Qt.AlignLeft

            if stn_y > self.height() / 2:
                sgn_y = -1
                label_y = stn_y - fm.height()
                align |= QtCore.Qt.AlignBottom
            else:
                sgn_y = 1
                label_y = stn_y
                align |= QtCore.Qt.AlignTop

            self.stn_readout.setText(self.stn_names[idxs[stn_idx]])
            self.stn_readout.move(label_x + sgn_x * label_offset, label_y + sgn_y * label_offset)
            self.stn_readout.setFixedWidth(fm.width(self.stn_names[idxs[stn_idx]]))
            self.stn_readout.setAlignment(align)
            self.setCursor(QtCore.Qt.PointingHandCursor)
        else:
            self.stn_readout.setText("")
            self.stn_readout.setFixedWidth(0)
            self.stn_readout.move(self.width(), self.height())
            self.stn_readout.setAlignment(QtCore.Qt.AlignLeft)
            self.unsetCursor()



import numpy as np
import numpy.ma as ma
import numpy.testing as npt
import sharppy.sharptab.interp as interp
from sharppy.sharptab.utils import vec2comp
from sharppy.sharptab.profile import Profile
import test_profile as tp


prof = tp.TestProfile().prof


def test_pres():
    input_z = 1000.
    correct_p = 903.8343884049208
    returned_p = interp.pres(prof, input_z)
    npt.assert_almost_equal(returned_p, correct_p)

    input_z = [1000., 3000., 6000.]
    correct_p = np.asarray([903.834388405, 710.02200544, 482.16636819])
    returned_p = interp.pres(prof, input_z)
    npt.assert_almost_equal(returned_p, correct_p)


def test_hght():
    input_p = 900
    correct_z = 1036.0022240687013
    returned_z = interp.hght(prof, input_p)
    npt.assert_almost_equal(returned_z, correct_z)

    input_p = [900, 800, 600, 400]
    correct_z = np.asarray([1036.0022240687013, 2020.3374024,
                            4330.6739399, 7360.])
    returned_z = interp.hght(prof, input_p)
    npt.assert_almost_equal(returned_z, correct_z)


def test_temp():
    input_p = 900
    correct_t = 14.589978853117268
    returned_t = interp.temp(prof, input_p)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = [900, 800, 600, 400]
    correct_t = np.asarray([14.589978853117268, 8.3624187,
                            -7.48619696, -29.7])
    returned_t = interp.temp(prof, input_p)
    npt.assert_almost_equal(returned_t, correct_t)


def test_dwpt():
    input_p = 900
    correct_t = 11.52599445805832
    returned_t = interp.dwpt(prof, input_p)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = [900, 800, 600, 400]
    correct_t = np.asarray([11.52599445805832, 6.07174951,
                            -13.34348479, -35.7])
    returned_t = interp.dwpt(prof, input_p)
    npt.assert_almost_equal(returned_t, correct_t)


def test_vtmp():
    input_p = 900
    correct_v = 16.24810167687504
    returned_v = interp.vtmp(prof, input_p)
    npt.assert_almost_equal(returned_v, correct_v, decimal=2)

    input_p = [900, 800, 600, 400]
    correct_v = np.asarray([16.24810167687504, 9.62205272,
                            -7.11816357, -29.63245875])
    returned_v = interp.vtmp(prof, input_p)
    npt.assert_almost_equal(returned_v, correct_v, decimal=2)


def test_components():
    input_p = 900
    correct_u, correct_v = -5.53976475, 20.6746835
    correct = [correct_u, correct_v]
    returned = interp.components(prof, input_p)
    npt.assert_almost_equal(returned, correct)

    input_p = [900, 800, 600, 400]
    correct_u = np.asarray([-5.53976475, 5.95267234,
                            23.10783339, 42.])
    correct_v = np.asarray([20.6746835, 15.54170573,
                            -9.37502817, 0])
    correct = [correct_u, correct_v]
    returned = interp.components(prof, input_p)
    npt.assert_almost_equal(returned, correct)


def test_vec():
    input_p = 900
    correct_wdir, correct_wspd = 165., 21.40400736
    correct = [correct_wdir, correct_wspd]
    returned = interp.vec(prof, input_p)
    npt.assert_almost_equal(returned, correct)

    input_p = [900, 800, 600, 400]
    correct_wdir = np.asarray([165., 200.95747979,
                               292.08277812, 270.])
    correct_wspd = np.asarray([21.40400736, 16.64268383,
                               24.93718343, 42.])
    correct = [correct_wdir, correct_wspd]
    returned = interp.vec(prof, input_p)
    npt.assert_almost_equal(returned, correct)


def test_to_agl():
    input_z = 1000.
    correct_agl = 643.0
    returned_agl = interp.to_agl(prof, input_z)
    npt.assert_almost_equal(returned_agl, correct_agl)

    input_z = [1000., 3000., 6000.]
    correct_agl = [643., 2643., 5643.]
    returned_agl = interp.to_agl(prof, input_z)
    npt.assert_almost_equal(returned_agl, correct_agl)


def test_to_msl():
    input_z = 1000.
    correct_agl = 1357.
    returned_agl = interp.to_msl(prof, input_z)
    npt.assert_almost_equal(returned_agl, correct_agl)

    input_z = [1000., 3000., 6000.]
    correct_agl = [1357., 3357., 6357]
    returned_agl = interp.to_msl(prof, input_z)
    npt.assert_almost_equal(returned_agl, correct_agl)






import pytest
matplotlib = pytest.importorskip('matplotlib')
#import matplotlib
matplotlib.use('Agg')
import sharppy.io.spc_decoder as spc_decoder
import sharppy.plot.skew as skew
import sharppy.sharptab.profile as profile
from matplotlib.ticker import ScalarFormatter, MultipleLocator
from matplotlib.collections import LineCollection
import matplotlib.transforms as transforms
from matplotlib import gridspec
import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime

"""
    Unit tests to test to see if decoders work on different file types
"""
files = ['examples/data/14061619.OAX']
 
def test_plotting():
   
    dec = spc_decoder.SPCDecoder(files[0])
    profs = dec.getProfiles()
    stn_id = dec.getStnId()

    print(profs) 
    all_profs = profs._profs
    print(all_profs)
    prof = all_profs[''][0]
    dates = profs._dates
    print(dates)
    prof = profile.create_profile(pres=prof.pres, hght=prof.hght, tmpc=prof.tmpc, dwpc=prof.dwpc, wspd=prof.wspd, \
                                      wdir=prof.wdir, strictQC=False, profile='convective', date=dates[0])
    time = dates[0]
    location = "OAX"
    pb_plot=1050
    pt_plot=100
    dp_plot=10
    plevs_plot = np.arange(pb_plot,pt_plot-1,-dp_plot)
    # Open up the text file with the data in columns (e.g. the sample OAX file distributed with SHARPpy)
    title = time.strftime('%Y%m%d/%H%M') + ' ' + location + '   (Observed)'

    # Set up the figure in matplotlib.
    fig = plt.figure(figsize=(9, 8))
    gs = gridspec.GridSpec(4,4, width_ratios=[1,5,1,1])
    ax = plt.subplot(gs[0:3, 0:2], projection='skewx')
    skew.draw_title(ax, title)
    skew.draw_dry_adiabats(ax)
    skew.draw_mixing_ratio_lines(ax)
    skew.draw_moist_adiabats(ax)
    skew.draw_heights(ax, prof)
    skew.draw_effective_inflow_layer(ax, prof)
    
    ax.grid(True)
    plt.grid(True)

    # Plot the background variables
    presvals = np.arange(1000, 0, -10)

    ax.semilogy(prof.tmpc[~prof.tmpc.mask], prof.pres[~prof.tmpc.mask], 'r', lw=2)
    ax.semilogy(prof.dwpc[~prof.dwpc.mask], prof.pres[~prof.dwpc.mask], 'g', lw=2)
    ax.semilogy(prof.vtmp[~prof.dwpc.mask], prof.pres[~prof.dwpc.mask], 'r--')
    ax.semilogy(prof.wetbulb[~prof.dwpc.mask], prof.pres[~prof.dwpc.mask], 'c-')

    # Plot the parcel trace, but this may fail.  If it does so, inform the user.
    try:
        ax.semilogy(prof.mupcl.ttrace, prof.mupcl.ptrace, 'k--')
    except:
        print("Couldn't plot parcel traces...")
    
    skew.plot_sig_levels(ax, prof)

    # Highlight the 0 C and -20 C isotherms.
    l = ax.axvline(0, color='b', ls='--')
    l = ax.axvline(-20, color='b', ls='--')

    # Disables the log-formatting that comes with semilogy
    ax.yaxis.set_major_formatter(ScalarFormatter())
    ax.set_yticks(np.linspace(100,1000,10))
    ax.set_ylim(1050,100)

    # Plot the hodograph data.
    inset_axes = skew.draw_hodo_inset(ax, prof)
    skew.plotHodo(inset_axes, prof.hght, prof.u, prof.v, color='r')
    #inset_axes.text(srwind[0], srwind[1], 'RM', color='r', fontsize=8)
    #inset_axes.text(srwind[2], srwind[3], 'LM', color='b', fontsize=8)

    # Draw the wind barbs axis and everything that comes with it.
    ax.xaxis.set_major_locator(MultipleLocator(10))
    ax.set_xlim(-50,50)
    ax2 = plt.subplot(gs[0:3,2])
    ax3 = plt.subplot(gs[3,0:3])
    skew.plot_wind_axes(ax2)
    skew.plot_wind_barbs(ax2, prof.pres, prof.u, prof.v)
    gs.update(left=0.05, bottom=0.05, top=0.95, right=1, wspace=0.025)
    #plt.show()


test_plotting()



from .SHARPPYMAIN.sharppy.sharptab.profile import create_profile
from .SHARPPYMAIN.sharppy.sharptab.interp import *
from .SHARPPYMAIN.sharppy.sharptab.winds import *
from .SHARPPYMAIN.sharppy.sharptab.utils import comp2vec 
from .SHARPPYMAIN.sharppy.sharptab.params import *

from ecape_parcel.ecape_calc import calc_ecape

import numpy as np
import numpy.ma as ma
import warnings
import metpy.calc as mpcalc
from metpy.units import units



#########################################################
#            SOUNDERPY SPYCALC FUNCTIONS                #
# (C) KYLE J GILLETT, CENTRAL MICHIGAN UNIVERSTIY, 2024 #
#########################################################



class sounding_params:


    def __init__(self, clean_data, storm_motion='right_moving'):
            self.clean_data = clean_data 
            self.storm_motion = storm_motion

            
    ######################################################################################################
    ### CALC FUNCTION -- COMPUTE VALUES ###
    ######################################################################################################        
    def calc(self):
        
        def calculate_height(pressure, pressure_at_sea_level=1013.25):
            # Constants
            R = 287.05  # Specific gas constant for dry air in J/(kg K)
            T0 = 288.15  # Standard temperature at sea level in K
            g = 9.8  # Acceleration due to gravity in m/s^2
            # Calculate height using the hypsometric equation
            height = (R * T0 / g) * np.log(pressure_at_sea_level / pressure)
            return height
    
        def find_nearest(array, value):
            array = np.asarray(array)
            nearest_idx = (np.abs(array - value)).argmin()
            return nearest_idx
    
    

        general = {}
        kinem  = {} 
        thermo = {}
        intrp  = {}

        # declare easy variable names for reuse from `clean_data`
        # first, check for correct units: 
        if self.clean_data['T'].units != 'degree_Celsius':
            self.clean_data['T'] =  self.clean_data['T'].to(units.degC)
        if self.clean_data['Td'].units != 'degree_Celsius':
            self.clean_data['Td'] = self.clean_data['Td'].to(units.degC)
        
        T   = self.clean_data['T']
        Td  = self.clean_data['Td']
        p   = self.clean_data['p']
        z   = self.clean_data['z']
        u   = self.clean_data['u']
        v   = self.clean_data['v']
        wd  = mpcalc.wind_direction(u, v)
        ws  = mpcalc.wind_speed(u, v)

        # compute useful variables and add them to a new dict of 
        # data that this function will return 
        if self.clean_data['site_info']['site-elv'] == 9999:
            general['elevation'] = z[0].m
        else:
            general['elevation'] = int(self.clean_data['site_info']['site-elv'])


        general['sfc_pressure']  = (p[0].m*(1-(0.0065*(general['elevation']))/(T[0].m+(0.0065*(general['elevation']))+273.15))**-5.257)*units.hPa
        general['wet_bulb']      = mpcalc.wet_bulb_temperature(p, T, Td)
        general['rel_humidity']  = mpcalc.relative_humidity_from_dewpoint(T, Td)*100
        general['spec_humidity'] = (mpcalc.specific_humidity_from_dewpoint(p, Td)*1000)*units.g/units.kg
        general['mix_ratio']     = mpcalc.mixing_ratio_from_specific_humidity(general['spec_humidity'])
        general['virt_temp']     = mpcalc.virtual_temperature(T, general['mix_ratio'])
        general['theta']         = mpcalc.potential_temperature(p, T)
        general['theta_e']       = mpcalc.equivalent_potential_temperature(p, T, Td)
        try:
            general['pwat']      = mpcalc.precipitable_water(p, Td).to('in')
            if general['pwat'] < 0:
                general['pwat'] = ma.masked
        except:
            general['pwat'] = ma.masked
        ##########################################   



        def interpolate(var,hgts,step):
            levels=np.arange(0,np.max(hgts),step)
            varinterp=np.zeros(len(levels))
            for i in range(0,len(levels)):
                try:
                    lower=np.where(hgts-levels[i]<=0,hgts-levels[i],-np.inf).argmax()
                    varinterp[i]=(((var[lower+1]-var[lower])/(hgts[lower+1]-hgts[lower]))*(levels[i]-hgts[lower])+var[lower])
                except IndexError:
                    # Handle the case where the index is out of bounds
                    varinterp[i] = np.nan 
            return varinterp 

        resolution=100

        rhINTRP = general['rel_humidity'].m[np.isnan(z)==False]
        mrINTRP = general['mix_ratio'].m[np.isnan(z)==False]
        thetaINTRP  = general['theta'].m[np.isnan(z)==False]
        thetaeINTRP = general['theta_e'].m[np.isnan(z)==False]
        pINTRP  = p.m[np.isnan(z)==False]
        tINTRP  = T.m[np.isnan(z)==False]
        uINTRP  = u.m[np.isnan(z)==False]
        vINTRP  = v.m[np.isnan(z)==False]
        zINTRP  = z.m[np.isnan(z)==False]
        zINTRP  = zINTRP-zINTRP[0]
        # Interpolation, add to intrp dict
        intrp['rhINTRP'] = interpolate(rhINTRP,zINTRP,resolution)
        intrp['tINTRP']  = interpolate(tINTRP,zINTRP,resolution)
        intrp['mrINTRP'] = interpolate(mrINTRP,zINTRP,resolution)
        intrp['thetaINTRP']  = interpolate(thetaINTRP,zINTRP,resolution)
        intrp['thetaeINTRP'] = interpolate(thetaeINTRP,zINTRP,resolution)
        intrp['pINTRP'] = interpolate(pINTRP,zINTRP,resolution)
        intrp['uINTRP'] = interpolate(uINTRP,zINTRP,resolution)
        intrp['vINTRP'] = interpolate(vINTRP,zINTRP,resolution)
        intrp['zINTRP'] = interpolate(zINTRP,zINTRP,resolution)

        hgt_lvls = {}
        hgt_var_list = ['h0', 'h05', 'h1', 'h15', 'h2', 'h25', 
                        'h3', 'h35', 'h4', 'h45', 'h5', 'h55', 
                        'h6', 'h65', 'h7', 'h75', 'h8', 'h85', 
                        'h9', 'h10', 'h11', 'h12', 'h13', 'h14']
        hgt_val_list = [0.,500.,1000.,1500.,2000.,2500.,3000.,3500.,
                        4000.,4500.,5000.,5500.,6000.,6500.,7000.,
                        7500.,8000.,8500.,9000.,10000.,11000.,
                        12000.,13000.,14000.]

        for var, val, in zip(hgt_var_list, hgt_val_list):
            try:
                hgt_lvls[var] = np.where(intrp['zINTRP']==val)[0][0]
            except:
                pass
        intrp['hgt_lvls'] = hgt_lvls


        # FREEZING POINT CALCULATION---           
        try: 
            frz_pt_index = intrp['tINTRP'].tolist().index(list(filter(lambda i: i <= 0, intrp['tINTRP'].tolist()))[0])
            general['frz_pt_p'] = intrp['pINTRP'][frz_pt_index]*units.hPa
            general['frz_pt_z'] = intrp['zINTRP'][frz_pt_index]*units.m
        except IndexError:
            general['frz_pt_p'] = ma.masked
            general['frz_pt_z'] = ma.masked
            warnings.warn("This sounding does not have a freezing point (not enough data)", Warning)
            pass



        ###################################################################
        ### SHARPPY VARIABLES CALC ###
        ###################################################################
        # create sharppy profile object
        prof = create_profile(profile='default',pres=p.m, hght=z.m, tmpc=T.m, dwpc=Td.m, wspd=ws, wdir=wd, 
                                      missing=-9999, strictQC=False)

        ####################################################################
        ### THERMODYNAMIC VALUES ###
        ####################################################################
        #--- SURFACE BASED PARCEL PROPERTIES ---#
        # ---------------------------------------------------------------
        sbpcl = parcelx(prof, flag=1, pres=p[0].m, tmpc=T[0].m, dwpc=Td[0].m)
        thermo['sbT_trace'] = sbpcl.ttrace
        thermo['sbP_trace'] = sbpcl.ptrace
        thermo['sbZ_trace'] = general['elevation'] + calculate_height(thermo['sbP_trace'])
        # compute heights
        thermo['sb_lcl_p'] = sbpcl.lclpres
        thermo['sb_lcl_z'] = sbpcl.lclhght 
        thermo['sb_lfc_p'] = sbpcl.lfcpres
        thermo['sb_lfc_z'] = sbpcl.lfchght
        thermo['sb_el_p']  = sbpcl.elpres  
        thermo['sb_el_z']  = sbpcl.elhght
        thermo['sb_el_T']  = temp(prof, pres(prof, thermo['sb_el_z']))
        thermo['sbcape']   = sbpcl.bplus
        thermo['sbcin']    = sbpcl.bminus
        thermo['sb3cape']  = sbpcl.b3km
        thermo['sb6cape']  = sbpcl.b6km

        #--- MOST UNSTABLE PARCEL PROPERTIES ---#
        # ---------------------------------------------------------------
        mupcl_p, mupcl_T, mupcl_Td = mpcalc.most_unstable_parcel(p, T, Td)[0:3]
        mupcl = parcelx(prof, flag=3, pres=mupcl_p.m, tmpc=mupcl_T.m, dwpc=mupcl_Td.m)
        thermo['muT_trace'] = mupcl.ttrace
        thermo['muP_trace'] = mupcl.ptrace
        thermo['muZ_trace'] = general['elevation'] + calculate_height(thermo['muP_trace'])
        # compute heights
        thermo['mu_lcl_p'] = mupcl.lclpres 
        thermo['mu_lcl_z'] = mupcl.lclhght 
        thermo['mu_lfc_p'] = mupcl.lfcpres 
        thermo['mu_lfc_z'] = mupcl.lfchght
        thermo['mu_el_p']  = mupcl.elpres
        thermo['mu_el_z']  = mupcl.elhght
        thermo['mu_el_T']  = temp(prof, pres(prof, thermo['mu_el_z']))
        thermo['mucape']   = mupcl.bplus
        thermo['mucin']    = mupcl.bminus
        thermo['mu3cape']  = mupcl.b3km
        thermo['mu6cape']  = mupcl.b6km

        #--- MIXED LAYER PARCEL PROPERTIES ---#
        # ---------------------------------------------------------------
        mlpcl_p, mlpcl_T, mlpcl_Td = mpcalc.mixed_parcel(p, T, Td, bottom=p[0], depth=50*units.hPa, interpolate=True)[0:3]
        mlpcl = parcelx(prof, flag=5, pres=mlpcl_p.m, tmpc=mlpcl_T.m, dwpc=mlpcl_Td.m)
        thermo['mlT_trace'] = mlpcl.ttrace
        thermo['mlP_trace'] = mlpcl.ptrace
        # compute heights
        thermo['ml_lcl_p'] = mlpcl.lclpres
        thermo['ml_lcl_z'] = mlpcl.lclhght
        thermo['ml_lfc_p'] = mlpcl.lfcpres
        thermo['ml_lfc_z'] = mlpcl.lfchght
        thermo['ml_el_p']  = mlpcl.elpres
        thermo['ml_el_z']  = mlpcl.elhght
        thermo['ml_el_T']  = temp(prof, pres(prof, thermo['ml_el_z']))
        thermo['mlcape']   = mlpcl.bplus
        thermo['mlcin']    = mlpcl.bminus
        thermo['ml3cape']  = mlpcl.b3km
        thermo['ml6cape']  = mlpcl.b6km

        #--- NORMALIZED CAPE ---#
        # ---------------------------------------------------------------
        thermo['mu_ncape'] = np.round(thermo['mucape'] / ((thermo['mu_el_z'] - thermo['mu_lfc_z'])),1)
        thermo['sb_ncape'] = np.round(thermo['sbcape'] / ((thermo['sb_el_z'] - thermo['sb_lfc_z'])),1)
        thermo['ml_ncape'] = np.round(thermo['mlcape'] / ((thermo['ml_el_z'] - thermo['ml_lfc_z'])),1) 

        #--- DOWNDRAFT CAPE ---#
        # ---------------------------------------------------------------
        thermo['dcape'], thermo['dparcel_T'], thermo['dparcel_p'] = dcape(prof)

        
        # ENTRAINING CAPE NOW LOCATED BELOW KINEMATICS SECTION SO 
        # CALCULATED STORM MOTION IS INCLUDED IN ECAPE CALCULATIONS
        
        
        
        
        
        
        #--- PBL ---#
        # ---------------------------------------------------------------
        thermo['pbl_top'] = pbl_top(prof) 

        #--- DGZ ---#
        # ---------------------------------------------------------------
        thermo['dgz'] = dgz(prof)

        #--- HGZ ---#
        # ---------------------------------------------------------------
        thermo['hgz'] = hgz(prof)

        #--- LAPSE RATES ---#
        # ---------------------------------------------------------------
        thermo['lr_36km'] = lapse_rate(prof, 3000, 6000, pres=False)
        thermo['lr_03km'] = lapse_rate(prof, 0, 3000, pres=False) 
        thermo['lr_max']  = max_lapse_rate(prof, lower=0, upper=6000, 
                                         interval=250, depth=2000)
        
        #--- TEMPERATURE ADVECTION ---#
        # ---------------------------------------------------------------
        #returns temp_adv (C/hr) array and 2D array of top and bottom bounds of temp advection layer (mb)
        thermo['temp_adv'] = inferred_temp_adv(prof, lat=self.clean_data['site_info']['site-latlon'][0])
        ####################################################################
        
        



        ####################################################################
        ### KINEMATIC VALUES ###
        ####################################################################
        #--- LAPSE RATES ---#
        # ---------------------------------------------------------------   
        kinem['eil'] = effective_inflow_layer(prof, ecape=100, ecinh=- 250)
        kinem['eil_z'] = [to_agl(prof, hght(prof, kinem['eil'][0])), to_agl(prof, hght(prof, kinem['eil'][1]))]

        #--- STORM MOTION ---#
        # ---------------------------------------------------------------   
        sm_rm = bunkers_storm_motion(prof)[0:2] 
        sm_lm = bunkers_storm_motion(prof)[2:4] 
        sm_mw = mean_wind(prof, pbot=850, ptop=250)

        if ma.is_masked(sm_rm[0]) == True:
            try:
                kinem['sm_rm'], kinem['sm_lm'], kinem['sm_mw'] = mpcalc.bunkers_storm_motion(p, u, v, z)
                kinem['sm_rm'], kinem['sm_lm'], kinem['sm_mw'] = kinem['sm_rm'].m, kinem['sm_lm'].m, kinem['sm_mw'].m
            except:
                kinem['sm_rm'], kinem['sm_lm'], kinem['sm_mw'] = ma.masked, ma.masked, ma.masked
                warnings.warn("Bunkers Storm Motion could not be computed for this sounding (not enough data)", Warning)
                pass
        elif np.isnan(sm_rm[0]) == True:
            kinem['sm_rm'], kinem['sm_lm'], kinem['sm_mw'] = ma.masked, ma.masked, ma.masked
            warnings.warn("Bunkers Storm Motion could not be computed for this sounding (not enough data)", Warning)
        else:
            kinem['sm_rm'], kinem['sm_lm'], kinem['sm_mw'] = sm_rm, sm_lm, sm_mw
            
            
        #--- USER DEFINED SM PT ---#
        # ---------------------------------------------------------------  
        if str(type(self.storm_motion)) == "<class 'list'>":
            kinem['sm_u'], kinem['sm_v'] = mpcalc.wind_components(self.storm_motion[1]*units.kts, self.storm_motion[0]*units.deg)
            kinem['sm_u'], kinem['sm_v'] = kinem['sm_u'].m, kinem['sm_v'].m
            
        else:
            
            if ma.is_masked(kinem['sm_rm']) == False:
                if self.storm_motion.casefold() in ['rm', 'right', 'right_moving', 'right_mover']:
                    kinem['sm_u'] = kinem['sm_rm'][0]
                    kinem['sm_v'] = kinem['sm_rm'][1]

                elif self.storm_motion.casefold() in ['lm', 'left', 'left_moving', 'right_mover']:
                    kinem['sm_u'] = kinem['sm_lm'][0]
                    kinem['sm_v'] = kinem['sm_lm'][1]

                elif self.storm_motion.casefold() in ['mw', 'mean', 'mean_wind', 'right_mover']:
                    kinem['sm_u'] = kinem['sm_mw'][0]
                    kinem['sm_v'] = kinem['sm_mw'][1]

            else:
                kinem['sm_u'] = ma.masked
                kinem['sm_v'] = ma.masked
            
                
        #--- DEVIANT TORNADO MOTION ---#
        # --------------------------------------------------------------- 
        if ma.is_masked(kinem['sm_u']) == False:
            p300m = pres(prof, to_msl(prof, 300.))
            sfc = prof.pres[prof.sfc]
            mw_u_300, mw_v_300 = mean_wind(prof, pbot=sfc, ptop=p300m)
            kinem['dtm'] = ((kinem['sm_u']+mw_u_300)/2, (kinem['sm_v']+mw_v_300)/2)
        else:
            kinem['dtm'] = ma.masked
          
        
        #--- CORFIDI MSC MOTION ---#
        #----------------------------------------------------------------
        kinem['mcs'] = corfidi_mcs_motion(prof)


        #--- BULK SHEAR ---#
        # ---------------------------------------------------------------    
        sfc = prof.pres[prof.sfc]
        p500m = pres(prof, to_msl(prof, 500.))
        p1km = pres(prof, to_msl(prof, 1000.))
        p3km = pres(prof, to_msl(prof, 3000.))
        p6km = pres(prof, to_msl(prof, 6000.))
        p9km = pres(prof, to_msl(prof, 9000.))

        shear_0_to_500  = wind_shear(prof, pbot=sfc, ptop=p500m)
        shear_0_to_1000 = wind_shear(prof, pbot=sfc, ptop=p1km)
        shear_1_to_3000 = wind_shear(prof, pbot=p1km, ptop=p3km)
        shear_3_to_6000 = wind_shear(prof, pbot=p3km, ptop=p6km)
        shear_6_to_9000 = wind_shear(prof, pbot=p6km, ptop=p9km)

        kinem['shear_0_to_500']  = comp2vec(shear_0_to_500[0], shear_0_to_500[1])[1]
        kinem['shear_0_to_1000'] = comp2vec(shear_0_to_1000[0], shear_0_to_1000[1])[1]
        kinem['shear_1_to_3000'] = comp2vec(shear_1_to_3000[0], shear_1_to_3000[1])[1]
        kinem['shear_3_to_6000'] = comp2vec(shear_3_to_6000[0], shear_3_to_6000[1])[1]
        kinem['shear_6_to_9000'] = comp2vec(shear_6_to_9000[0], shear_6_to_9000[1])[1]
            
        #--- SRH ---#
        # ---------------------------------------------------------------
        if ma.is_masked(kinem['sm_u']) == False:
            kinem['srh_0_to_500']  = helicity(prof, 0, 500.,     stu = kinem['sm_u'], stv = kinem['sm_v'])[0]
            kinem['srh_0_to_1000'] = helicity(prof, 0, 1000.,    stu = kinem['sm_u'], stv = kinem['sm_v'])[0]
            kinem['srh_1_to_3000'] = helicity(prof, 1000, 3000., stu = kinem['sm_u'], stv = kinem['sm_v'])[0]
            kinem['srh_3_to_6000'] = helicity(prof, 3000, 6000., stu = kinem['sm_u'], stv = kinem['sm_v'])[0]
            kinem['srh_6_to_9000'] = helicity(prof, 6000, 9000., stu = kinem['sm_u'], stv = kinem['sm_v'])[0]
        else:
            kinem['srh_0_to_500']  = ma.masked
            kinem['srh_0_to_1000'] = ma.masked
            kinem['srh_1_to_3000'] = ma.masked
            kinem['srh_3_to_6000'] = ma.masked
            kinem['srh_6_to_9000'] = ma.masked
            warnings.warn("Storm Relative Helicity could not be computed for this sounding (no valid storm motion)", Warning)


        #--- SRW LAYERS ---#
        # --------------------------------------------------------------- 
        if ma.is_masked(kinem['sm_u']) == False:
            def calc_srw_layer(prof, level):
                u, v = components(prof, p=pres(prof, to_msl(prof, level)))
                sru = u - kinem['sm_u']
                srv = v - kinem['sm_v']
                return mpcalc.wind_speed(sru*units.kts, srv*units.kts)
            try:
                kinem['srw_0_to_500']  = (calc_srw_layer(prof, 500)  + calc_srw_layer(prof, 0))/2
                kinem['srw_0_to_1000'] = (calc_srw_layer(prof, 1000) + calc_srw_layer(prof, 0))/2
                kinem['srw_1_to_3000'] = (calc_srw_layer(prof, 3000) + calc_srw_layer(prof, 1000))/2
                kinem['srw_3_to_6000'] = (calc_srw_layer(prof, 6000) + calc_srw_layer(prof, 3000))/2
                kinem['srw_6_to_9000'] = (calc_srw_layer(prof, 9000) + calc_srw_layer(prof, 6000))/2
            except:
                kinem['srw_0_to_500']  = ma.masked
                kinem['srw_0_to_1000'] = ma.masked
                kinem['srw_1_to_3000'] = ma.masked
                kinem['srw_3_to_6000'] = ma.masked
                kinem['srw_6_to_9000'] = ma.masked
                pass
        else:
            kinem['srw_0_to_500']  = ma.masked
            kinem['srw_0_to_1000'] = ma.masked
            kinem['srw_1_to_3000'] = ma.masked
            kinem['srw_3_to_6000'] = ma.masked
            kinem['srw_6_to_9000'] = ma.masked
            warnings.warn("Storm Relative Wind could not be computed for this sounding (no valid storm motion)", Warning)
            
        #--- SRV ---#
        # ---------------------------------------------------------------
        # adopted from Sam Brandt (2022)    
        if ma.is_masked(kinem['sm_u']) == False: 
            # CONVERT TO m/s (uses `sm_u, sm_v` calculated above)
            u_ms = (intrp['uINTRP']/1.94384)
            v_ms = (intrp['vINTRP']/1.94384)
            sm_u_ms = ((intrp['uINTRP'] - kinem['sm_u'])/1.94384)
            sm_v_ms = ((intrp['vINTRP'] - kinem['sm_v'])/1.94384)

            # INTEROPLATED SRW (send back to knots)
            srw = mpcalc.wind_speed(sm_u_ms*units('m/s'), sm_v_ms*units('m/s'))
            kinem['srw'] = (srw.m*1.94384)

            # SHEAR COMPONENTS FOR VORT CALC
            # calc example = change in u over change in z
            dudz = (u_ms[2::]-u_ms[0:-2]) / (intrp['zINTRP'][2::]-intrp['zINTRP'][0:-2])
            dvdz = (v_ms[2::]-v_ms[0:-2]) / (intrp['zINTRP'][2::]-intrp['zINTRP'][0:-2])
            dudz = np.insert(dudz,0,dudz[0])
            dudz = np.insert(dudz,-1,dudz[-1])
            dvdz = np.insert(dvdz,0,dvdz[0])
            dvdz = np.insert(dvdz,-1,dvdz[-1])
            # Shear magnitude, 
            shear=(np.sqrt(dudz**2+dvdz**2)+0.0000001)
            # Vorticity components
            uvort=-dvdz
            vvort=dudz
            # Total horizontal vorticity
            kinem['vort'] = np.sqrt(uvort**2 + vvort**2)
            # Total streamwise vorticity
            kinem['swv'] = abs((sm_u_ms*uvort+sm_v_ms*vvort)/(np.sqrt(sm_u_ms**2+sm_v_ms**2)))
            # Streamwiseness fraction
            kinem['swv_perc'] = (kinem['swv']/shear)*100

            # layer average streamwiseness and total streamwise vorticity
            kinem['swv_perc_0_to_500']  = np.mean(kinem['swv_perc'][0:5])
            kinem['swv_perc_0_to_1000'] = np.mean(kinem['swv_perc'][0:10])
            kinem['swv_perc_1_to_3000'] = np.mean(kinem['swv_perc'][10:30])
            kinem['swv_perc_3_to_6000'] = np.mean(kinem['swv_perc'][30:60])
            kinem['swv_perc_6_to_9000'] = np.mean(kinem['swv_perc'][60:90])
            kinem['swv_0_to_500']       = np.mean(kinem['swv'][0:5])
            kinem['swv_0_to_1000']      = np.mean(kinem['swv'][0:10])
            kinem['swv_1_to_3000']      = np.mean(kinem['swv'][10:30])
            kinem['swv_3_to_6000']      = np.mean(kinem['swv'][30:60])
            kinem['swv_6_to_9000']      = np.mean(kinem['swv'][60:90])
        else:
            kinem['srw']                = ma.masked
            kinem['swv']                = ma.masked
            kinem['swv_perc']           = ma.masked
            kinem['vort']               = ma.masked
            kinem['swv_perc_0_to_500']  = ma.masked
            kinem['swv_perc_0_to_1000'] = ma.masked
            kinem['swv_perc_1_to_3000'] = ma.masked
            kinem['swv_perc_3_to_6000'] = ma.masked
            kinem['swv_perc_6_to_9000'] = ma.masked
            kinem['swv_0_to_500']       = ma.masked
            kinem['swv_0_to_1000']      = ma.masked
            kinem['swv_1_to_3000']      = ma.masked
            kinem['swv_3_to_6000']      = ma.masked
            kinem['swv_6_to_9000']      = ma.masked
            warnings.warn("Streamwise Vorticity could not be computed for this sounding (no valid storm motion)", Warning)
        
        
        #--- SIMPLE ENTRAINING CAPE ---#
        # ---------------------------------------------------------------
        try: 
            thermo['mu_ecape'] = calc_ecape(z, p, T, general['spec_humidity'], u, v, 'most_unstable', storm_motion='user_defined', u_sm=kinem['sm_u']*units.kts, v_sm=kinem['sm_v']*units.kts).m
        except: 
            thermo['mu_ecape'] = ma.masked
            warnings.warn("MU-ECAPE could not be computed for this sounding (calculation error)", Warning)
            pass
        
        try:
            thermo['ml_ecape'] = calc_ecape(z, p, T, general['spec_humidity'], u, v, 'mixed_layer', storm_motion='user_defined', u_sm=kinem['sm_u']*units.kts, v_sm=kinem['sm_v']*units.kts).m
        except: 
            thermo['ml_ecape'] = ma.masked
            warnings.warn("ML-ECAPE could not be computed for this sounding (calculation error)", Warning)
            pass
        
        try:
            thermo['sb_ecape'] = calc_ecape(z, p, T, general['spec_humidity'], u, v, 'surface_based', storm_motion='user_defined', u_sm=kinem['sm_u']*units.kts, v_sm=kinem['sm_v']*units.kts).m
        except:
            thermo['sb_ecape'] = ma.masked
            warnings.warn("SB-ECAPE could not be computed for this sounding (calculation error)", Warning)
            pass
        
        
        
        #--- RH & MIXRAT WITH HEIGHT ---#
        # ---------------------------------------------------------------
        general['rh_0_500']  = np.mean(intrp['rhINTRP'][0:5])
        general['rh_0_1000'] = np.mean(intrp['rhINTRP'][0:10])
        general['rh_1_3000'] = np.mean(intrp['rhINTRP'][10:30])
        general['rh_3_6000'] = np.mean(intrp['rhINTRP'][30:60])
        general['rh_6_9000'] = np.mean(intrp['rhINTRP'][60:90])
        

        general['w_0_500']  = np.mean(intrp['mrINTRP'][0:5]) 
        general['w_0_1000'] = np.mean(intrp['mrINTRP'][0:10]) 
        general['w_1_3000'] = np.mean(intrp['mrINTRP'][10:30]) 
        general['w_3_6000'] = np.mean(intrp['mrINTRP'][30:60]) 
        general['w_6_9000'] = np.mean(intrp['mrINTRP'][60:90]) 
        
        return general, thermo, kinem, intrp 

    
    
    ######################################################################################################
    ### PRINT FUNCTION -- PRINT A FEW VALUES ###
    ######################################################################################################  
    def print_vals(self):
        thermo = self.calc()[1]
        kinem  = self.calc()[2]
        print(' ')
        print(f'> THERMODYNAMICS --------------------------------------------- ')
        print(f"--- SBCAPE: {np.round(thermo['sbcape'], 1)} | MUCAPE: {np.round(thermo['mucape'], 1)} | MLCAPE: {np.round(thermo['mlcape'], 1)} | MUECAPE: {np.round(thermo['mu_ecape'], 1)}")
        print(f"--- MU 0-3: {np.round(thermo['mu3cape'], 1)} | MU 0-6: {np.round(thermo['mu6cape'], 1)} | SB 0-3: {np.round(thermo['sb3cape'], 1)} | SB 0-6: {np.round(thermo['sb6cape'], 1)}")
        print(' ')
        print(f'> KINEMATICS ------------------------------------------------- ')
        print(f"--- 0-500 SRW: {np.round(kinem['srw_0_to_500'], 1)} | 0-500 SWV: {np.round(kinem['swv_0_to_500'],3)} | 0-500 SHEAR: {np.round(kinem['shear_0_to_500'], 1)} | 0-500 SRH: {np.round(kinem['srh_0_to_500'], 1)}")
        print(f"--- 1-3km SRW: {np.round(kinem['srw_1_to_3000'], 1)} | 1-3km SWV: {np.round(kinem['swv_1_to_3000'],3)} | 1-3km SHEAR: {np.round(kinem['shear_1_to_3000'], 1)} | | 1-3km SRH: {np.round(kinem['srh_1_to_3000'], 1)}")
        print(f'============================================================== ')
        print(" ")











# VAD Profile Calc functions 

class vad_params:
    
    def __init__(self, vad_data, storm_motion='right_moving'):
            self.vad_data = vad_data 
            self.storm_motion = storm_motion
            
    
    ######################################################################################################
    ### CALC FUNCTION -- COMPUTE VALUES ###
    ######################################################################################################        
    def calc(self):
    
        def find_nearest(array, value):
            array = np.asarray(array)
            nearest_idx = (np.abs(array - value)).argmin()
            return nearest_idx
    
    
        kinem  = {} 
        intrp  = {}

        z   = self.vad_data['z']
        u   = self.vad_data['u']
        v   = self.vad_data['v']
        wd  = mpcalc.wind_direction(u*units.kts, v*units.kts)
        ws  = mpcalc.wind_speed(u*units.kts, v*units.kts)

        # compute useful variables and add them to a new dict of 
        # data that this function will return 
        ##########################################   



        def interpolate(var,hgts,step):
            levels=np.arange(0,np.max(hgts),step)
            varinterp=np.zeros(len(levels))
            for i in range(0,len(levels)):
                try:
                    lower=np.where(hgts-levels[i]<=0,hgts-levels[i],-np.inf).argmax()
                    varinterp[i]=(((var[lower+1]-var[lower])/(hgts[lower+1]-hgts[lower]))*(levels[i]-hgts[lower])+var[lower])
                except IndexError:
                    # Handle the case where the index is out of bounds
                    varinterp[i] = np.nan 
            return varinterp 

        resolution=100
        
        uINTRP  = u[np.isnan(z)==False]
        vINTRP  = v[np.isnan(z)==False]
        zINTRP  = z[np.isnan(z)==False]
        zINTRP  = zINTRP-zINTRP[0]
        
        # Interpolation, add to intrp dict
        intrp['uINTRP'] = interpolate(uINTRP,zINTRP,resolution)
        intrp['vINTRP'] = interpolate(vINTRP,zINTRP,resolution)
        intrp['zINTRP'] = interpolate(zINTRP,zINTRP,resolution)

        hgt_lvls = {}
        hgt_var_list = ['h0', 'h05', 'h1', 'h15', 'h2', 'h25', 
                        'h3', 'h35', 'h4', 'h45', 'h5', 'h55', 
                        'h6', 'h65', 'h7', 'h75', 'h8', 'h85', 
                        'h9']
        
        hgt_val_list = [0.,500.,1000.,1500.,2000.,2500.,3000.,3500.,
                        4000.,4500.,5000.,5500.,6000.,6500.,7000.,
                        7500.,8000.,8500.,9000.]

        for var, val, in zip(hgt_var_list, hgt_val_list):
            try:
                hgt_lvls[var] = np.where(intrp['zINTRP']==val)[0][0]
            except:
                pass
        intrp['hgt_lvls'] = hgt_lvls
        

        ####################################################################
        ### KINEMATIC VALUES ###
        ####################################################################
        
        # THE FOLLOWING FUNCTIONS ARE CONTRUBUTED BY SCOTT THOMAS, NWS GRAND RAPIDS 
        # -------------------------------------------------------------------------  

        # CALC WIND VECTORS
        def calc_vector(u_comp, v_comp):
              mag = np.sqrt(u_comp**2 + v_comp**2)
              dir = np.rad2deg(np.arctan2(u_comp, v_comp)) % 360
              return mag, dir

            
        # CALC LAYER SHEAR
        def calc_shear(u_layer, v_layer, bottom, top, z):
              layer_top = np.where(z == (top))[0][0]
              layer_bot = np.where(z == (bottom))[0][0]
              u_shr = u_layer[layer_top] - u_layer[layer_bot]
              v_shr = v_layer[layer_top] - v_layer[layer_bot]
              shrmag = np.hypot(u_shr, v_shr)
              return shrmag

        
        # CALC MEAN_WIND COMPONENTS 
        def calc_meanwind(u_layer, v_layer, z, top):
              layer_top = np.where(z == (top))[0][0]
              mean_u = np.mean(u_layer[:layer_top])
              mean_v = np.mean(v_layer[:layer_top])
              return mean_u, mean_v

        
        # CALC BUNKERS STORM MOTION
        def calc_bunkers(u_layer, v_layer, z):
              layer_top = np.where(z == (6000))[0][0]
              mean_u = np.mean(u_layer[:layer_top])
              mean_v = np.mean(v_layer[:layer_top])

              layer_top = np.where(z == (6000))[0][0]
              u_shr = u_layer[layer_top] - u_layer[0]
              v_shr = v_layer[layer_top] - v_layer[0]

              dev = 7.5 * 1.94

              dev_amnt = dev / np.hypot(u_shr, v_shr)
              rmu = mean_u + (dev_amnt * v_shr)
              rmv = mean_v - (dev_amnt * u_shr)
              lmu = mean_u - (dev_amnt * v_shr)
              lmv = mean_v + (dev_amnt * u_shr)
              return rmu, rmv, lmu, lmv
            
            
    
        # CALC CORFIDI UP SHEAR & DOWN SHEAR STORM MOTION
        def calc_corfidi(u_layer, v_layer, z, u_mean, v_mean):
              llj_top = np.where(z == (1500))[0][0]
              llj_u = u_layer[:llj_top]
              llj_v = v_layer[:llj_top]

              mag, dir = calc_vector(llj_u, llj_v)
              max=0
              i=0
              for a in mag:
                if mag[i] >= mag[i-1]:
                  max = i

              u_max = llj_u[i]
              v_max = llj_v[i]

              corfidi_up_u = u_mean - u_max
              corfidi_up_v =  v_mean - v_max

              corfidi_down_u = u_mean + corfidi_up_u
              corfidi_down_v = v_mean + corfidi_up_v

              return [corfidi_up_u, corfidi_up_v, corfidi_down_u, corfidi_down_v]

        
        # CALC STORM RELATIVE HELICITY
        def calc_srh(u_comp, v_comp, z, depth, bottom, storm_u, storm_v):
            srh = mpcalc.storm_relative_helicity(height=z, u=u_comp, v=v_comp, depth=depth,
                                           bottom=bottom, storm_u=storm_u, storm_v=storm_v)[0]
            return srh.m
                
            
        # CALC STORM RELATIVE WIND 
        def calc_srw_layer(u, v, bottom, top):
            layer_top = np.where(z == (top))[0][0]
            layer_bot = np.where(z == (bottom))[0][0]
            
            top_sru = u[layer_top] - kinem['sm_u']
            top_srv = v[layer_top] - kinem['sm_v']
            
            bot_sru = u[layer_bot] - kinem['sm_u']
            bot_srv = v[layer_bot] - kinem['sm_v']
            
            mean_srw = (mpcalc.wind_speed(top_sru*units.kts, top_srv*units.kts) + mpcalc.wind_speed(bot_sru*units.kts, bot_srv*units.kts))/2
            
            return mean_srw.m
        
                
        
            
        #--- BUNKERS STORM MOTION ---#
        # --------------------------------------------------------------- 
        kinem['sm_rm'] = calc_bunkers(u, v, z)[0:2]
        kinem['sm_lm'] = calc_bunkers(u, v, z)[2:4]
        kinem['sm_mw'] = calc_meanwind(u, v, z, 6000)
       
        
        #--- USER DEFINED SM PT ---#
        # ---------------------------------------------------------------  
        if str(type(self.storm_motion)) == "<class 'list'>":
            kinem['sm_u'], kinem['sm_v'] = mpcalc.wind_components(self.storm_motion[1]*units.kts, self.storm_motion[0]*units.deg)
            kinem['sm_u'], kinem['sm_v'] = kinem['sm_u'].m, kinem['sm_v'].m
            
        else:
            
            if ma.is_masked(kinem['sm_rm']) == False:
                if self.storm_motion.casefold() in ['rm', 'right', 'right_moving', 'right_mover']:
                    kinem['sm_u'] = kinem['sm_rm'][0]
                    kinem['sm_v'] = kinem['sm_rm'][1]

                elif self.storm_motion.casefold() in ['lm', 'left', 'left_moving', 'right_mover']:
                    kinem['sm_u'] = kinem['sm_lm'][0]
                    kinem['sm_v'] = kinem['sm_lm'][1]

                elif self.storm_motion.casefold() in ['mw', 'mean', 'mean_wind', 'right_mover']:
                    kinem['sm_u'] = kinem['sm_mw'][0]
                    kinem['sm_v'] = kinem['sm_mw'][1]

            else:
                kinem['sm_u'] = ma.masked
                kinem['sm_v'] = ma.masked
                
                
                
        #--- DEVIANT TORNADO MOTION ---#
        # --------------------------------------------------------------- 
        if ma.is_masked(kinem['sm_u']) == False:
            mw_u_300, mw_v_300 = calc_meanwind(u, v, z, 300)
            kinem['dtm'] = ((kinem['sm_u']+mw_u_300)/2, (kinem['sm_v']+mw_v_300)/2)
        else:
            kinem['dtm'] = ma.masked
          
        
        
        #--- CORFIDI MSC MOTION ---#
        #----------------------------------------------------------------
        kinem['mcs'] = calc_corfidi(u, v, z, 
                                    calc_meanwind(u, v, z, 6000)[0],
                                    calc_meanwind(u, v, z, 6000)[1])


        
        #--- BULK SHEAR ---#
        # ---------------------------------------------------------------  
        kinem['shear_0_to_500']  = calc_shear(u, v, 0, 500, z)
        kinem['shear_0_to_1000'] = calc_shear(u, v, 0, 1000, z)
        kinem['shear_1_to_3000'] = calc_shear(u, v, 1000, 3000, z)
        kinem['shear_3_to_6000'] = calc_shear(u, v, 3000, 6000, z)
        kinem['shear_6_to_9000'] = calc_shear(u, v, 6000, 9000, z)
            
            
            
        #--- SRH ---#
        # ---------------------------------------------------------------
        if ma.is_masked(kinem['sm_u']) == False:
            kinem['srh_0_to_500']  = calc_srh(u_comp=u*units.kts, v_comp=v*units.kts, z=z*units.m, 
                                              bottom=None, depth=500*units.m,  
                                              storm_u=kinem['sm_u']*units.kts, storm_v=kinem['sm_v']*units.kts)
            kinem['srh_0_to_1000'] = calc_srh(u_comp=u*units.kts, v_comp=v*units.kts, z=z*units.m, 
                                              bottom=None, depth=1000*units.m, 
                                              storm_u=kinem['sm_u']*units.kts, storm_v=kinem['sm_v']*units.kts)
            kinem['srh_1_to_3000'] = calc_srh(u_comp=u*units.kts, v_comp=v*units.kts, z=z*units.m, 
                                              bottom=1000*units.m, depth=3000*units.m, 
                                              storm_u=kinem['sm_u']*units.kts, storm_v=kinem['sm_v']*units.kts)
            kinem['srh_3_to_6000'] = calc_srh(u_comp=u*units.kts, v_comp=v*units.kts, z=z*units.m, 
                                              bottom=3000*units.m, depth=6000*units.m, 
                                              storm_u=kinem['sm_u']*units.kts, storm_v=kinem['sm_v']*units.kts)
            kinem['srh_6_to_9000'] = calc_srh(u_comp=u*units.kts, v_comp=v*units.kts, z=z*units.m, 
                                              bottom=6000*units.m, depth=9000*units.m, 
                                              storm_u=kinem['sm_u']*units.kts, storm_v=kinem['sm_v']*units.kts)
        else:
            kinem['srh_0_to_500']  = ma.masked
            kinem['srh_0_to_1000'] = ma.masked
            kinem['srh_1_to_3000'] = ma.masked
            kinem['srh_3_to_6000'] = ma.masked
            kinem['srh_6_to_9000'] = ma.masked
            warnings.warn("Storm Relative Helicity could not be computed for this sounding (no valid storm motion)", Warning)


            
        #--- SRW LAYERS ---#
        # --------------------------------------------------------------- 
        if ma.is_masked(kinem['sm_u']) == False:
            kinem['srw_0_to_500']  = calc_srw_layer(u, v, 0, 500)
            kinem['srw_0_to_1000'] = calc_srw_layer(u, v, 0, 1000)
            kinem['srw_1_to_3000'] = calc_srw_layer(u, v, 1000, 3000)
            kinem['srw_3_to_6000'] = calc_srw_layer(u, v, 3000, 6000)
            kinem['srw_6_to_9000'] = calc_srw_layer(u, v, 6000, 9000)

        
        
        #--- SRV ---#
        # ---------------------------------------------------------------
        # adopted from Sam Brandt (2022)    
        if ma.is_masked(kinem['sm_u']) == False: 
            # CONVERT TO m/s (uses `sm_u, sm_v` calculated above)
            u_ms = (intrp['uINTRP']/1.94384)
            v_ms = (intrp['vINTRP']/1.94384)
            sm_u_ms = ((intrp['uINTRP'] - kinem['sm_u'])/1.94384)
            sm_v_ms = ((intrp['vINTRP'] - kinem['sm_v'])/1.94384)

            # INTEROPLATED SRW (send back to knots)
            srw = mpcalc.wind_speed(sm_u_ms*units('m/s'), sm_v_ms*units('m/s'))
            kinem['srw'] = (srw.m*1.94384)

            # SHEAR COMPONENTS FOR VORT CALC
            # calc example = change in u over change in z
            dudz = (u_ms[2::]-u_ms[0:-2]) / (intrp['zINTRP'][2::]-intrp['zINTRP'][0:-2])
            dvdz = (v_ms[2::]-v_ms[0:-2]) / (intrp['zINTRP'][2::]-intrp['zINTRP'][0:-2])
            dudz = np.insert(dudz,0,dudz[0])
            dudz = np.insert(dudz,-1,dudz[-1])
            dvdz = np.insert(dvdz,0,dvdz[0])
            dvdz = np.insert(dvdz,-1,dvdz[-1])
            # Shear magnitude, 
            shear=(np.sqrt(dudz**2+dvdz**2)+0.0000001)
            # Vorticity components
            uvort=-dvdz
            vvort=dudz
            # Total horizontal vorticity
            kinem['vort'] = np.sqrt(uvort**2 + vvort**2)
            # Total streamwise vorticity
            kinem['swv'] = abs((sm_u_ms*uvort+sm_v_ms*vvort)/(np.sqrt(sm_u_ms**2+sm_v_ms**2)))
            # Streamwiseness fraction
            kinem['swv_perc'] = (kinem['swv']/shear)*100

            # layer average streamwiseness and total streamwise vorticity
            kinem['swv_perc_0_to_500']  = np.mean(kinem['swv_perc'][0:5])
            kinem['swv_perc_0_to_1000'] = np.mean(kinem['swv_perc'][0:10])
            kinem['swv_perc_1_to_3000'] = np.mean(kinem['swv_perc'][10:30])
            kinem['swv_perc_3_to_6000'] = np.mean(kinem['swv_perc'][30:60])
            kinem['swv_perc_6_to_9000'] = np.mean(kinem['swv_perc'][60:90])
            kinem['swv_0_to_500']       = np.mean(kinem['swv'][0:5])
            kinem['swv_0_to_1000']      = np.mean(kinem['swv'][0:10])
            kinem['swv_1_to_3000']      = np.mean(kinem['swv'][10:30])
            kinem['swv_3_to_6000']      = np.mean(kinem['swv'][30:60])
            kinem['swv_6_to_9000']      = np.mean(kinem['swv'][60:90])
        else:
            kinem['srw']                = ma.masked
            kinem['swv']                = ma.masked
            kinem['swv_perc']           = ma.masked
            kinem['vort']               = ma.masked
            kinem['swv_perc_0_to_500']  = ma.masked
            kinem['swv_perc_0_to_1000'] = ma.masked
            kinem['swv_perc_1_to_3000'] = ma.masked
            kinem['swv_perc_3_to_6000'] = ma.masked
            kinem['swv_perc_6_to_9000'] = ma.masked
            kinem['swv_0_to_500']       = ma.masked
            kinem['swv_0_to_1000']      = ma.masked
            kinem['swv_1_to_3000']      = ma.masked
            kinem['swv_3_to_6000']      = ma.masked
            kinem['swv_6_to_9000']      = ma.masked
            warnings.warn("Streamwise Vorticity could not be computed for this sounding (no valid storm motion)", Warning)
        
        return kinem, intrp 


import sharppy.sharptab.profile as profile
from sharppy.io.decoder import getDecoders
import sys

def decode(filename):

    for decname, deccls in getDecoders().items():
        try:
            dec = deccls(filename)
            break
        except:
            dec = None
            continue

    if dec is None:
        raise IOError("Could not figure out the format of '%s'!" % filename)

    # Returns the set of profiles from the file that are from the "Profile" class.
    profs = dec.getProfiles()
    stn_id = dec.getStnId()

    return profs, stn_id

def test_url():
    path = 'examples/data/14061619.OAX'
    # Sys.argv[1] should be the URL to the file that is being tested.
    profs, stn_id = decode(path)
    print((profs._profs))

    for k in profs._profs.keys():
        all_prof = profs._profs[k]
        dates = profs._dates
        for i in range(len(all_prof)):
            prof = all_prof[i]
            new_prof = profile.create_profile(pres=prof.pres, hght=prof.hght, tmpc=prof.tmpc, dwpc=prof.dwpc, wspd=prof.wspd, \
                                              wdir=prof.wdir, strictQC=False, profile='convective', date=dates[i])
            #for key in dir(new_prof):
            #    print((key, getattr(new_prof,key)))
    
    
    print(new_prof.mupcl.bplus)




from . import thermo, utils, interp, constants
import numpy as np

## Routines implemented in Python by Greg Blumberg - CIMMS and Kelton Halbert (OU SoM)
## wblumberg@ou.edu, greg.blumberg@noaa.gov, kelton.halbert@noaa.gov, keltonhalbert@ou.edu

def fosberg(prof):
    '''
        The Fosberg Fire Weather Index
        Adapted from code donated by Rich Thompson - NOAA Storm Prediction Center

        Description:
        The FWI (Fire Weather Index) is defined by a quantitative model that provides
        a nonlinear filter of meteorological data which results in a linear relationship
        between the combined meteorological variables of relative humidity and wind speed,
        and the behavior of wildfires. Thus the index deals with only the weather conditions,
        not the fuels. Several sets of conditions have been defined by Fosberg (Fosberg, 1978)
        to apply this to fire weather management. The upper limits have been set to give an
        index value of 100 if the moisture content is zero and the wind is 30 mph. 

        Thus, the numbers range from 0 to 100 and if any number is larger than 100, it is set back to 100. 
        The index can be used to measure changes in fire weather conditions. Over several years of use, 
        Fosberg index values of 50 or greater generally appear significant on a national scale.
        The SPC fire weather verification scheme uses the Fosberg Index, but with a check for
        both temperature (60F) and adjective fire danger rating (3-High, 4-Very High, 5-Extreme).

        Description Source - http://www.spc.noaa.gov/exper/firecomp/INFO/fosbinfo.html

        WARNING: This function has not been fully tested.

        Parameters
        ----------
        prof : profile object
            Profile object

        Returns
        -------
        param : number
            Fosberg Fire Weather Index

    '''
    tmpf = thermo.ctof(prof.tmpc[prof.get_sfc()])
    fmph = utils.KTS2MPH(prof.wspd[prof.get_sfc()])

    rh = thermo.relh(prof.pres[prof.sfc], prof.tmpc[prof.sfc], prof.dwpc[prof.sfc])
    if (rh <= 10):
        em = 0.03229 + 0.281073*rh - 0.000578*rh*tmpf
    elif (10 < rh <= 50):
        em = 2.22749 + 0.160107*rh - 0.014784*tmpf
    else:
        em = 21.0606 + 0.005565*rh*rh - 0.00035*rh*tmpf - 0.483199*rh

    em30 = em/30
    u_sq = fmph*fmph
    fmdc = 1 - 2*em30 + 1.5*em30*em30 - 0.5*em30*em30*em30

    param = (fmdc*np.sqrt(1+u_sq))/0.3002

    return param

def haines_height(prof):
    '''
        Haines Index Height calculation
        
        Calculates the appropriate height category(Low/Mid/High) given the 
        the lowest height in the sounding. 
        
        Adapted from S-591 course
        Added by Nickolai Reimer (NWS Billings, MT)
                
        Parameters
        ----------
        prof : profile object
            Profile object

        Returns
        -------
        param : number
            the Haines Index Height

    '''
    sfc_elevation = prof.hght[prof.sfc]
    
    # Haines low elevation below 1000 ft / 305 m
    if sfc_elevation < 305:
        return constants.HAINES_LOW
    
    # Haines mid elevation between 1000 ft / 305 m and 3000 ft / 914 m
    elif 305 <= sfc_elevation and sfc_elevation <= 914:
        return constants.HAINES_MID
    
    # Haines high elevation above 3000 ft / 914 m
    else:
        return constants.HAINES_HIGH
        
def haines_low(prof):
    '''
        Haines Index Low Elevation calculation
        
        Calculates the Haines Index(Lower Atmosphere Severity Index)
        using the lower elevation parmeters, used below 1000ft or 305 m.
        
        Pressure levels 950 mb and 850 mb
        Dewpoint depression at 850 mb
        
        Lapse Rate Term
        ---------------
        1 : < 4C
        2 : 4C to 7C
        3 : > 7C
        
        Dewpoint Depression Term
        ------------------------
        1 : < 6C
        2 : 6C to 9C
        3 : > 9C
        
        Adapted from S-591 course 
        Added by Nickolai Reimer (NWS Billings, MT)
        
        Parameters
        ----------
        prof : profile object
            Profile object

        Returns
        -------
        param : number
            the Haines Index low

    '''
    
    tp1  = interp.temp(prof, 950)
    tp2  = interp.temp(prof, 850)
    tdp2 = interp.dwpt(prof, 850)
    
    if utils.QC(tp1) and utils.QC(tp2) and utils.QC(tdp2):
        lapse_rate = tp1 - tp2
        dewpoint_depression = tp2 - tdp2
        
        if lapse_rate < 4:
            a = 1
        elif 4 <= lapse_rate and lapse_rate <= 7:
            a = 2
        else:
            a = 3
        
        if dewpoint_depression < 6:
            b = 1
        elif 6 <= dewpoint_depression and dewpoint_depression <= 9:
            b = 2
        else:
            b = 3
        return a + b
    else:
        return constants.MISSING

def haines_mid(prof):
    '''
        Haines Index Mid Elevation calculation
        
        Calculates the Haines Index(Lower Atmosphere Severity Index)
        using the middle elevation parmeters, used 
        between 1000 ft or 305 m and 3000 ft or 914 m.
        
        Pressure levels 850 mb and 700 mb
        Dewpoint depression at 850 mb
        
        Lapse Rate Term
        ---------------
        1 : < 6C
        2 : 6C to 10C
        3 : > 10C
        
        Dewpoint Depression Term
        ------------------------
        1 : < 6C
        2 : 6C to 12C
        3 : > 12C
        
        Adapted from S-591 course
        Added by Nickolai Reimer (NWS Billings, MT)
        
        Parameters
        ----------
        prof : profile object
            Profile object

        Returns
        -------
        param : number
            the Haines Index mid

    '''
    
    tp1  = interp.temp(prof, 850)
    tp2  = interp.temp(prof, 700)
    tdp1 = interp.dwpt(prof, 850)
    
    if utils.QC(tp1) and utils.QC(tp2) and utils.QC(tdp1):
        lapse_rate = tp1 - tp2
        dewpoint_depression = tp1 - tdp1
        
        if lapse_rate < 6:
            a = 1
        elif 6 <= lapse_rate and lapse_rate <= 10:
            a = 2
        else:
            a = 3
        
        if dewpoint_depression < 6:
            b = 1
        elif 6 <= dewpoint_depression and dewpoint_depression <= 12:
            b = 2
        else:
            b = 3
        return a + b
    else:
        return constants.MISSING

def haines_high(prof):
    '''
        Haines Index High Elevation calculation
        
        Calculates the Haines Index(Lower Atmosphere Severity Index)
        using the higher elevation parmeters, used above 3000ft or 914 m.
        
        Pressure levels 700 mb and 500 mb
        Dewpoint depression at 700 mb
        
        Lapse Rate Term
        ---------------
        1 : < 18C
        2 : 18C to 21C
        3 : > 21C
        
        Dewpoint Depression Term
        ------------------------
        1 : < 15C
        2 : 15C to 20C
        3 : > 20C
        
        Adapted from S-591 course
        Added by Nickolai Reimer (NWS Billings, MT)
        
        Parameters
        ----------
        prof : profile object
            Profile object

        Returns
        -------
        param : number
            the Haines Index high

    '''
    tp1  = interp.temp(prof, 700)
    tp2  = interp.temp(prof, 500)
    tdp1 = interp.dwpt(prof, 700)
    
    if utils.QC(tp1) and utils.QC(tp2) and utils.QC(tdp1):
        lapse_rate = tp1 - tp2
        dewpoint_depression = tp1 - tdp1
        
        if lapse_rate < 18:
            a = 1
        elif 18 <= lapse_rate and lapse_rate <= 21:
            a = 2
        else:
            a = 3
        
        if dewpoint_depression < 15:
            b = 1
        elif 15 <= dewpoint_depression and dewpoint_depression <= 20:
            b = 2
        else:
            b = 3
        return a + b
    else:
        return constants.MISSING





from qtpy.QtWidgets import *
import sys

# A short utility to use PySide to detect the screen resolution
# (for use in testing the SHARPpy GUI)

app = QApplication(sys.argv)
screen_rect = app.desktop().screenGeometry()
width, height = screen_rect.width(), screen_rect.height()
print("Screen Resolution:", str(width) + 'x' + str(height))



import numpy as np
import numpy.ma as ma
import numpy.testing as npt
import sharppy.sharptab.thermo as thermo
from sharppy.sharptab.constants import *


def test_ctof():
    # single pass
    input_c = 0
    correct_f = 32
    returned_f = thermo.ctof(input_c)
    npt.assert_almost_equal(returned_f, correct_f)

    # array_like pass
    input_c = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    input_c = np.asanyarray(input_c)
    correct_f = [32, 50, 68, 86, 104, 122, 140, 158, 176, 194, 212]
    correct_f = np.asanyarray(correct_f)
    returned_f = thermo.ctof(input_c)
    npt.assert_almost_equal(returned_f, correct_f)

    # single masked
    input_c = ma.masked
    correct_f = ma.masked
    returned_f = thermo.ctof(input_c)
    npt.assert_equal(type(returned_f), type(correct_f))

    # array_like pass
    inds = [0, 5, 7]
    input_c = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    input_c = np.ma.asanyarray(input_c)
    correct_f = [32, 50, 68, 86, 104, 122, 140, 158, 176, 194, 212]
    correct_f = np.ma.asanyarray(correct_f)
    input_c[inds] = ma.masked
    correct_f[inds] = ma.masked
    returned_f = thermo.ctof(input_c)
    npt.assert_almost_equal(returned_f, correct_f)


def test_ftoc():
    # single pass
    input_f = 32
    correct_c = 0
    returned_c = thermo.ftoc(input_f)
    npt.assert_almost_equal(returned_c, correct_c)

    # array_like pass
    input_f = [32, 50, 68, 86, 104, 122, 140, 158, 176, 194, 212]
    input_f = np.asanyarray(input_f)
    correct_c = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    correct_c = np.asanyarray(correct_c)
    returned_c = thermo.ftoc(input_f)
    npt.assert_almost_equal(returned_c, correct_c)

    # single masked
    input_f = ma.masked
    correct_c = ma.masked
    returned_c = thermo.ftoc(input_f)
    npt.assert_equal(type(returned_c), type(correct_c))

    # array_like pass
    inds = [0, 5, 7]
    input_f = [32, 50, 68, 86, 104, 122, 140, 158, 176, 194, 212]
    input_f = np.ma.asanyarray(input_f)
    correct_c = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
    correct_c = np.ma.asanyarray(correct_c)
    input_f[inds] = ma.masked
    correct_c[inds] = ma.masked
    returned_c = thermo.ftoc(input_f)
    npt.assert_almost_equal(returned_c, correct_c)


def test_ktoc():
    # single pass
    input_k = 0
    correct_c = -273.15
    returned_c = thermo.ktoc(input_k)
    npt.assert_almost_equal(returned_c, correct_c)

    # array_like pass
    input_k = [0, 50, 100, 150, 200, 250, 300]
    input_k = np.asanyarray(input_k)
    correct_c = [-273.15, -223.15, -173.15, -123.15, -73.15, -23.15, 26.85]
    correct_c = np.asanyarray(correct_c)
    returned_c = thermo.ktoc(input_k)
    npt.assert_almost_equal(returned_c, correct_c)

    # single masked
    input_k = ma.masked
    correct_c = ma.masked
    returned_c = thermo.ktoc(input_k)
    npt.assert_equal(type(returned_c), type(correct_c))

    # array_like pass
    inds = [0, 2, 3]
    input_k = [0, 50, 100, 150, 200, 250, 300]
    input_k = np.ma.asanyarray(input_k)
    correct_c = [-273.15, -223.15, -173.15, -123.15, -73.15, -23.15, 26.85]
    correct_c = np.ma.asanyarray(correct_c)
    input_k[inds] = ma.masked
    correct_c[inds] = ma.masked
    returned_c = thermo.ktoc(input_k)
    npt.assert_almost_equal(returned_c, correct_c)


def test_ctok():
    # single pass
    input_c = -273.15
    correct_k = 0
    returned_k = thermo.ctok(input_c)
    npt.assert_almost_equal(returned_k, correct_k)

    # array_like pass
    input_c = [-273.15, -223.15, -173.15, -123.15, -73.15, -23.15, 26.85]
    input_c = np.asanyarray(input_c)
    correct_k = [0, 50, 100, 150, 200, 250, 300]
    correct_k = np.asanyarray(correct_k)
    returned_k = thermo.ctok(input_c)
    npt.assert_almost_equal(returned_k, correct_k)

    # single masked
    input_c = ma.masked
    correct_k = ma.masked
    returned_k = thermo.ctok(input_c)
    npt.assert_equal(type(returned_k), type(correct_k))

    # array_like pass
    inds = [0, 2, 3]
    input_c = [-273.15, -223.15, -173.15, -123.15, -73.15, -23.15, 26.85]
    input_c = np.ma.asanyarray(input_c)
    correct_k = [0, 50, 100, 150, 200, 250, 300]
    correct_k = np.ma.asanyarray(correct_k)
    input_c[inds] = ma.masked
    correct_k[inds] = ma.masked
    returned_k = thermo.ctok(input_c)
    npt.assert_almost_equal(returned_k, correct_k)


def test_ktof():
    # single pass
    input_k = 0
    correct_f = -459.67
    returned_f = thermo.ktof(input_k)
    npt.assert_almost_equal(returned_f, correct_f)

    # array_like pass
    input_k = [0, 50, 100, 150, 200, 250, 300]
    input_k = np.asanyarray(input_k)
    correct_f = [-459.67, -369.67, -279.67, -189.67, -99.67, -9.67, 80.33]
    correct_f = np.asanyarray(correct_f)
    returned_f = thermo.ktof(input_k)
    npt.assert_almost_equal(returned_f, correct_f)

    # single masked
    input_k = ma.masked
    correct_f = ma.masked
    returned_f = thermo.ktof(input_k)
    npt.assert_equal(type(returned_f), type(correct_f))

    # array_like pass
    inds = [0, 2, 3]
    input_k = [0, 50, 100, 150, 200, 250, 300]
    input_k = np.ma.asanyarray(input_k)
    correct_f = [-459.67, -369.67, -279.67, -189.67, -99.67, -9.67, 80.33]
    correct_f = np.ma.asanyarray(correct_f)
    input_k[inds] = ma.masked
    correct_f[inds] = ma.masked
    returned_f = thermo.ktof(input_k)
    npt.assert_almost_equal(returned_f, correct_f)


def test_ftok():
    # single pass
    input_f = -459.67
    correct_k = 0
    returned_k = thermo.ftok(input_f)
    npt.assert_almost_equal(returned_k, correct_k)

    # array_like pass
    input_f = [-459.67, -369.67, -279.67, -189.67, -99.67, -9.67, 80.33]
    input_f = np.asanyarray(input_f)
    correct_k = [0, 50, 100, 150, 200, 250, 300]
    correct_k = np.asanyarray(correct_k)
    returned_k = thermo.ftok(input_f)
    npt.assert_almost_equal(returned_k, correct_k)

    # single masked
    input_f = ma.masked
    correct_k = ma.masked
    returned_k = thermo.ftok(input_f)
    npt.assert_equal(type(returned_k), type(correct_k))

    # array_like pass
    inds = [0, 2, 3]
    input_f = [-459.67, -369.67, -279.67, -189.67, -99.67, -9.67, 80.33]
    input_f = np.ma.asanyarray(input_f)
    correct_k = [0, 50, 100, 150, 200, 250, 300]
    correct_k = np.ma.asanyarray(correct_k)
    input_f[inds] = ma.masked
    correct_k[inds] = ma.masked
    returned_k = thermo.ftok(input_f)
    npt.assert_almost_equal(returned_k, correct_k)


def test_theta():
    # single
    input_p = 940
    input_t = 5
    input_p2 = 1000.
    correct_theta = 9.961049492262532
    returned_theta = thermo.theta(input_p, input_t, input_p2)
    npt.assert_almost_equal(returned_theta, correct_theta)

    # array
    input_p = np.asarray([940, 850])
    input_t = np.asarray([5, 10])
    input_p2 = np.asarray([1000., 1000.])
    correct_theta = [9.961049492262532, 23.457812111895066]
    returned_theta = thermo.theta(input_p, input_t, input_p2)
    npt.assert_almost_equal(returned_theta, correct_theta)


def test_wobf():
    input_t = 10
    correct_c = 10.192034543230415
    returned_c = thermo.wobf(input_t)
    npt.assert_almost_equal(returned_c, correct_c)

    input_t = [10, 0, -10]
    input_t = np.asanyarray(input_t)
    correct_c = [10.192034543230415, 6.411053315058521, 3.8633154447163114]
    correct_c = np.asanyarray(correct_c)
    returned_c = thermo.wobf(input_t)
    npt.assert_almost_equal(returned_c, correct_c)


def test_lcltemp():
    input_t = 10
    input_td = 5
    correct_t = 3.89818375
    returned_t = thermo.lcltemp(input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

    input_t = np.asanyarray([20, 10, 0, -5])
    input_td = np.asanyarray([15, 8, -1, -10])
    correct_t = [13.83558375, 7.54631416, -1.21632173, -11.00791625]
    correct_t = np.asanyarray(correct_t)
    returned_t = thermo.lcltemp(input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)


def test_thalvl():
    input_theta = 10
    input_t = 5
    correct_p = 939.5475008003834
    returned_p = thermo.thalvl(input_theta, input_t)
    npt.assert_almost_equal(returned_p, correct_p)

    input_theta = np.asanyarray([5, 12, 25])
    input_t = np.asanyarray([5, 10, 0.])
    correct_p = [1000., 975.6659847653189, 736.0076986893786]
    correct_p = np.asanyarray(correct_p)
    returned_p = thermo.thalvl(input_theta, input_t)
    npt.assert_almost_equal(returned_p, correct_p)


def test_drylift():
    input_p = 950
    input_t = 30
    input_td = 25
    correct_p = 883.4367363248148
    correct_t = 23.77298375
    returned_p, returned_t = thermo.drylift(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_p, correct_p)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = np.asarray([950, 975, 1013, 900])
    input_t = np.asarray([30, 10, 22, 40])
    input_td = np.asarray([25, -10, 18, 0])
    correct_p = np.asarray([883.4367363248148, 716.8293994988512,
                            954.7701032005202, 504.72627541064145])
    correct_t = np.asarray([23.77298375, -13.822639999999996,
                            17.04965568, -7.6987199999999945])
    returned_p, returned_t = thermo.drylift(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_p, correct_p)
    npt.assert_almost_equal(returned_t, correct_t)


def test_satlift():
    input_p = 850
    input_thetam = 20
    correct_t = 13.712979340608157
    returned_t = thermo.satlift(input_p, input_thetam)
    npt.assert_almost_equal(returned_t, correct_t)


def test_wetlift():
    input_p = 700
    input_t = 15
    input_p2 = 100
    correct_t = -81.27400812504021
    returned_t = thermo.wetlift(input_p, input_t, input_p2)
    npt.assert_almost_equal(returned_t, correct_t)


def test_lifted():
    input_p = 950
    input_t = 30
    input_td = 25
    input_lev = 100
    correct_t = -79.05621246586672
    returned_t = thermo.lifted(input_p, input_t, input_td, input_lev)
    npt.assert_almost_equal(returned_t, correct_t)


def test_vappres():
    input_t = 25
    correct_p = 31.670078513287617
    returned_p = thermo.vappres(input_t)
    npt.assert_almost_equal(returned_p, correct_p)

    input_t = np.asanyarray([0, 5, 10, 15, 20, 25])
    correct_p = [6.107954896017587, 8.719365306196854, 12.2722963940349,
                 17.04353238898728, 23.37237439430437, 31.670078513287617]
    correct_p = np.asanyarray(correct_p)
    returned_p = thermo.vappres(input_t)
    npt.assert_almost_equal(returned_p, correct_p)


def test_mixratio():
    input_p = 950
    input_t = 25
    correct_w = 21.549675456205275
    returned_w = thermo.mixratio(input_p, input_t)
    npt.assert_almost_equal(returned_w, correct_w)

    input_p = np.asanyarray([1013, 1000, 975, 950, 900])
    input_t = np.asanyarray([26, 15, 20, 10, 10])
    correct_w = [21.448870702611913, 10.834359059077558, 15.346544211592512,
                 8.17527964576288, 8.633830400361578]
    correct_w = np.asanyarray(correct_w)
    returned_w = thermo.mixratio(input_p, input_t)
    npt.assert_almost_equal(returned_w, correct_w)


def test_temp_at_mixrat():
    input_w = 14
    input_p = 950
    correct_t = 18.25602418045935
    returned_t = thermo.temp_at_mixrat(input_w, input_p)
    npt.assert_almost_equal(returned_t, correct_t)

    input_w = np.asanyarray([14, 12, 10, 8])
    input_p = np.asanyarray([1013, 925, 850, 700])
    correct_t = [19.28487241829498, 15.451394956732088,
                 11.399344140651465, 5.290414578916341]
    correct_t = np.asanyarray(correct_t)
    returned_t = thermo.temp_at_mixrat(input_w, input_p)
    npt.assert_almost_equal(returned_t, correct_t)


def test_wetbulb():
    input_p = 950
    input_t = 5
    input_td = -10
    correct_t = -0.04811002960985089
    returned_t = thermo.wetbulb(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = 1013
    input_t = 5
    input_td = -10
    correct_t = 0.22705033380623352
    returned_t = thermo.wetbulb(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)


def test_thetaw():
    input_p = 925
    input_t = 7
    input_td = 3
    correct_t = 8.69793773351298
    returned_t = thermo.thetaw(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = 950
    input_t = 20
    input_td = 14
    correct_t = 18.21065472362592
    returned_t = thermo.thetaw(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)


def test_thetae():
    input_p = 925
    input_t = 7
    input_td = 3
    correct_t = 28.864469418729357
    returned_t = thermo.thetae(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = 950
    input_t = 20
    input_td = 14
    correct_t = 57.68849564698746
    returned_t = thermo.thetae(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)


def test_virtemp():
    input_p = 925
    input_t = 7
    input_td = 3
    correct_t = 7.873652724440433
    returned_t = thermo.virtemp(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = 950
    input_t = 20
    input_td = 14
    correct_t = 21.883780613639033
    returned_t = thermo.virtemp(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

def test_relh():
    input_p = 925
    input_t = 7
    input_td = 3
    correct_t = 75.6532482737457
    returned_t = thermo.relh(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)

    input_p = 950
    input_t = 20
    input_td = 14
    correct_t = 68.35638039676762
    returned_t = thermo.relh(input_p, input_t, input_td)
    npt.assert_almost_equal(returned_t, correct_t)









